r"""
Bulk2 single (A omic framework for bulk2single cell omic analysis)
"""

from ._bulk2single import Bulk2Single
from ._single2spatial import Single2Spatial
from ._bulktrajblend import BulkTrajBlend
from ._utils import bulk2single_plot_cellprop,bulk2single_plot_correlation


# import gpytorch
# from gpytorch.constraints import Positive, GreaterThan
# from gpytorch.lazy import DiagLazyTensor,BlockDiagLazyTensor,MatmulLazyTensor, InterpolatedLazyTensor, PsdSumLazyTensor, RootLazyTensor, KroneckerProductLazyTensor, lazify
# from gpytorch.utils.broadcasting import _mul_broadcast_shape
# from gpytorch.kernels.kernel import Kernel
# from gpytorch.kernels import MultitaskKernel
# from typing import Any
# from gpytorch.distributions import base_distributions, MultivariateNormal
# from gpytorch.functions import add_diag
# from gpytorch.likelihoods import Likelihood, _GaussianLikelihoodBase
# from gpytorch.utils.warnings import OldVersionWarning, GPInputWarning
# from gpytorch.likelihoods.noise_models import MultitaskHomoskedasticNoise, Noise, HomoskedasticNoise
# from gpytorch.models.exact_prediction_strategies import prediction_strategy
#
# import warnings
# import torch
# from torch import Tensor
# from gpytorch.mlls.marginal_log_likelihood import MarginalLogLikelihood
# # from mofapy2.core.distributions.myExactGP import myExactGP
#
# from gpytorch.models import ExactGP
# from gpytorch.models.gp import GP
# from gpytorch import settings
# from copy import deepcopy
#
#
# # Here we define a multitask GP model where
# # - the noise and scale parameter are coupled --> kernels do not see the raw noise parameter, liklihood is constructed based on K + likelihood.raw_noise I,
# # K is constructed by myMultitaskKernel
# # - the index kernel is constrained to a correlation matrix --> added a scaling factor to the IndexKernel
# # - likelihood is replaced by the ELBO --> the trace(Sigma^(-1) Cov(z)) term to the multitaskmarginal likelihood which is then optimized
#
# # Most of these classes are directly based on classed in gpytorch with minor changes as described above.
#
# # make sure it accepts the new likelihood
# class myExactGP(GP):
#
#     def __init__(self, train_inputs, train_targets, likelihood):
#         if train_inputs is not None and torch.is_tensor(train_inputs):
#             train_inputs = (train_inputs,)
#         if train_inputs is not None and not all(torch.is_tensor(train_input) for train_input in train_inputs):
#             raise RuntimeError("Train inputs must be a tensor, or a list/tuple of tensors")
#         if not isinstance(likelihood, _GaussianLikelihoodBase) and not isinstance(likelihood, _myGaussianLikelihoodBase) :
#             raise RuntimeError("myExactGP can only handle Gaussian likelihoods")
#
#         super(myExactGP, self).__init__()
#         if train_inputs is not None:
#             self.train_inputs = tuple(tri.unsqueeze(-1) if tri.ndimension() == 1 else tri for tri in train_inputs)
#             self.train_targets = train_targets
#         else:
#             self.train_inputs = None
#             self.train_targets = None
#         self.likelihood = likelihood
#
#         self.prediction_strategy = None
#
#     @property
#     def train_targets(self):
#         return self._train_targets
#
#     @train_targets.setter
#     def train_targets(self, value):
#         object.__setattr__(self, "_train_targets", value)
#
#     def _apply(self, fn):
#         if self.train_inputs is not None:
#             self.train_inputs = tuple(fn(train_input) for train_input in self.train_inputs)
#             self.train_targets = fn(self.train_targets)
#         return super(myExactGP, self)._apply(fn)
#
#     def local_load_samples(self, samples_dict, memo, prefix):
#         """
#         Replace the model's learned hyperparameters with samples from a posterior distribution.
#         """
#         # Pyro always puts the samples in the first batch dimension
#         num_samples = next(iter(samples_dict.values())).size(0)
#         self.train_inputs = tuple(tri.unsqueeze(0).expand(num_samples, *tri.shape) for tri in self.train_inputs)
#         self.train_targets = self.train_targets.unsqueeze(0).expand(num_samples, *self.train_targets.shape)
#         super().local_load_samples(samples_dict, memo, prefix)
#
#     def set_train_data(self, inputs=None, targets=None, strict=True):
#         """
#         Set training data (does not re-fit model hyper-parameters).
#
#         :param torch.Tensor inputs: The new training inputs.
#         :param torch.Tensor targets: The new training targets.
#         :param bool strict: (default True) If `True`, the new inputs and
#             targets must have the same shape, dtype, and device
#             as the current inputs and targets. Otherwise, any shape/dtype/device are allowed.
#         """
#         if inputs is not None:
#             if torch.is_tensor(inputs):
#                 inputs = (inputs,)
#             inputs = tuple(input_.unsqueeze(-1) if input_.ndimension() == 1 else input_ for input_ in inputs)
#             if strict:
#                 for input_, t_input in zip(inputs, self.train_inputs or (None,)):
#                     for attr in {"shape", "dtype", "device"}:
#                         expected_attr = getattr(t_input, attr, None)
#                         found_attr = getattr(input_, attr, None)
#                         if expected_attr != found_attr:
#                             msg = "Cannot modify {attr} of inputs (expected {e_attr}, found {f_attr})."
#                             msg = msg.format(attr=attr, e_attr=expected_attr, f_attr=found_attr)
#                             raise RuntimeError(msg)
#             self.train_inputs = inputs
#         if targets is not None:
#             if strict:
#                 for attr in {"shape", "dtype", "device"}:
#                     expected_attr = getattr(self.train_targets, attr, None)
#                     found_attr = getattr(targets, attr, None)
#                     if expected_attr != found_attr:
#                         msg = "Cannot modify {attr} of targets (expected {e_attr}, found {f_attr})."
#                         msg = msg.format(attr=attr, e_attr=expected_attr, f_attr=found_attr)
#                         raise RuntimeError(msg)
#             self.train_targets = targets
#         self.prediction_strategy = None
#
#     def get_fantasy_model(self, inputs, targets, **kwargs):
#         """
#         Returns a new GP model that incorporates the specified inputs and targets as new training data.
#
#         Using this method is more efficient than updating with `set_train_data` when the number of inputs is relatively
#         np.all, because any computed test-time caches will be updated in linear time rather than computed from scratch.
#
#         .. note::
#             If `targets` is a batch (e.g. `b x m`), then the GP returned from this method will be a batch mode GP.
#             If `inputs` is of the same (or lesser) dimension as `targets`, then it is assumed that the fantasy points
#             are the same for each target batch.
#
#         :param torch.Tensor inputs: (`b1 x ... x bk x m x d` or `f x b1 x ... x bk x m x d`) Locations of fantasy
#             observations.
#         :param torch.Tensor targets: (`b1 x ... x bk x m` or `f x b1 x ... x bk x m`) Labels of fantasy observations.
#         :return: An `myExactGP` model with `n + m` training examples, where the `m` fantasy examples have been added
#             and all test-time caches have been updated.
#         :rtype: ~gpytorch.models.myExactGP
#         """
#         if self.prediction_strategy is None:
#             raise RuntimeError(
#                 "Fantasy observations can only be added after making predictions with a model so that "
#                 "all test independent caches exist. Call the model on some data first!"
#             )
#
#         model_batch_shape = self.train_inputs[0].shape[:-2]
#
#         if self.train_targets.dim() > len(model_batch_shape) + 1:
#             raise RuntimeError("Cannot yet add fantasy observations to multitask GPs, but this is coming soon!")
#
#         if not isinstance(inputs, list):
#             inputs = [inputs]
#
#         inputs = [i.unsqueeze(-1) if i.ndimension() == 1 else i for i in inputs]
#
#         target_batch_shape = targets.shape[:-1]
#         input_batch_shape = inputs[0].shape[:-2]
#         tbdim, ibdim = len(target_batch_shape), len(input_batch_shape)
#
#         if not (tbdim == ibdim + 1 or tbdim == ibdim):
#             raise RuntimeError(
#                 f"Unsupported batch shapes: The target batch shape ({target_batch_shape}) must have either the "
#                 f"same dimension as or one more dimension than the input batch shape ({input_batch_shape})"
#             )
#
#         # Check whether we can properly broadcast batch dimensions
#         err_msg = (
#             f"Model batch shape ({model_batch_shape}) and target batch shape "
#             f"({target_batch_shape}) are not broadcastable."
#         )
#         _mul_broadcast_shape(model_batch_shape, target_batch_shape, error_msg=err_msg)
#
#         if len(model_batch_shape) > len(input_batch_shape):
#             input_batch_shape = model_batch_shape
#         if len(model_batch_shape) > len(target_batch_shape):
#             target_batch_shape = model_batch_shape
#
#         # If input has no fantasy batch dimension but target does, we can save memory and computation by not
#         # computing the covariance for each element of the batch. Therefore we don't expand the inputs to the
#         # size of the fantasy model here - this is done below, after the evaluation and fast fantasy update
#         train_inputs = [tin.expand(input_batch_shape + tin.shape[-2:]) for tin in self.train_inputs]
#         train_targets = self.train_targets.expand(target_batch_shape + self.train_targets.shape[-1:])
#
#         full_inputs = [
#             torch.cat([train_input, input.expand(input_batch_shape + input.shape[-2:])], dim=-2)
#             for train_input, input in zip(train_inputs, inputs)
#         ]
#         full_targets = torch.cat([train_targets, targets.expand(target_batch_shape + targets.shape[-1:])], dim=-1)
#
#         try:
#             fantasy_kwargs = {"noise": kwargs.pop("noise")}
#         except KeyError:
#             fantasy_kwargs = {}
#
#         full_output = super(myExactGP, self).__call__(*full_inputs, **kwargs)
#
#         # Copy model without copying training data or prediction strategy (since we'll overwrite those)
#         old_pred_strat = self.prediction_strategy
#         old_train_inputs = self.train_inputs
#         old_train_targets = self.train_targets
#         old_likelihood = self.likelihood
#         self.prediction_strategy = None
#         self.train_inputs = None
#         self.train_targets = None
#         self.likelihood = None
#         new_model = deepcopy(self)
#         self.prediction_strategy = old_pred_strat
#         self.train_inputs = old_train_inputs
#         self.train_targets = old_train_targets
#         self.likelihood = old_likelihood
#
#         new_model.likelihood = old_likelihood.get_fantasy_likelihood(**fantasy_kwargs)
#         new_model.prediction_strategy = old_pred_strat.get_fantasy_strategy(
#             inputs, targets, full_inputs, full_targets, full_output, **fantasy_kwargs
#         )
#
#         # if the fantasies are at the same points, we need to expand the inputs for the new model
#         if tbdim == ibdim + 1:
#             new_model.train_inputs = [fi.expand(target_batch_shape + fi.shape[-2:]) for fi in full_inputs]
#         else:
#             new_model.train_inputs = full_inputs
#         new_model.train_targets = full_targets
#
#         return new_model
#
#     def train(self, mode=True):
#         if mode:
#             self.prediction_strategy = None
#         return super(myExactGP, self).train(mode)
#
#     def _load_from_state_dict(
#         self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs
#     ):
#         self.prediction_strategy = None
#         super()._load_from_state_dict(
#             state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs
#         )
#
#     def __call__(self, *args, **kwargs):
#         train_inputs = list(self.train_inputs) if self.train_inputs is not None else []
#         inputs = [i.unsqueeze(-1) if i.ndimension() == 1 else i for i in args]
#
#         # Training mode: optimizing
#         if self.training:
#             if self.train_inputs is None:
#                 raise RuntimeError(
#                     "train_inputs, train_targets cannot be None in training mode. "
#                     "Call .eval() for prior predictions, or call .set_train_data() to add training data."
#                 )
#             if settings.debug.on():
#                 if not all(torch.equal(train_input, input) for train_input, input in zip(train_inputs, inputs)):
#                     raise RuntimeError("You must train on the training inputs!")
#             res = super().__call__(*inputs, **kwargs)
#             return res
#
#         # Prior mode
#         elif settings.prior_mode.on() or self.train_inputs is None or self.train_targets is None:
#             full_inputs = args
#             full_output = super(myExactGP, self).__call__(*full_inputs, **kwargs)
#             if settings.debug().on():
#                 if not isinstance(full_output, MultivariateNormal):
#                     raise RuntimeError("myExactGP.forward must return a MultivariateNormal")
#             return full_output
#
#         # Posterior mode
#         else:
#             if settings.debug.on():
#                 if all(torch.equal(train_input, input) for train_input, input in zip(train_inputs, inputs)):
#                     warnings.warn(
#                         "The input matches the stored training data. Did you forget to call model.train()?",
#                         GPInputWarning,
#                     )
#
#             # Get the terms that only depend on training data
#             if self.prediction_strategy is None:
#                 train_output = super().__call__(*train_inputs, **kwargs)
#
#                 # Create the prediction strategy for
#                 self.prediction_strategy = prediction_strategy(
#                     train_inputs=train_inputs,
#                     train_prior_dist=train_output,
#                     train_labels=self.train_targets,
#                     likelihood=self.likelihood,
#                 )
#
#             # Concatenate the input to the training input
#             full_inputs = []
#             batch_shape = train_inputs[0].shape[:-2]
#             for train_input, input in zip(train_inputs, inputs):
#                 # Make sure the batch shapes agree for training/test data
#                 if batch_shape != train_input.shape[:-2]:
#                     batch_shape = _mul_broadcast_shape(batch_shape, train_input.shape[:-2])
#                     train_input = train_input.expand(*batch_shape, *train_input.shape[-2:])
#                 if batch_shape != input.shape[:-2]:
#                     batch_shape = _mul_broadcast_shape(batch_shape, input.shape[:-2])
#                     train_input = train_input.expand(*batch_shape, *train_input.shape[-2:])
#                     input = input.expand(*batch_shape, *input.shape[-2:])
#                 full_inputs.append(torch.cat([train_input, input], dim=-2))
#
#             # Get the joint distribution for training/test data
#             full_output = super(myExactGP, self).__call__(*full_inputs, **kwargs)
#             if settings.debug().on():
#                 if not isinstance(full_output, MultivariateNormal):
#                     raise RuntimeError("myExactGP.forward must return a MultivariateNormal")
#             full_mean, full_covar = full_output.loc, full_output.lazy_covariance_matrix
#
#             # Determine the shape of the joint distribution
#             batch_shape = full_output.batch_shape
#             joint_shape = full_output.event_shape
#             tasks_shape = joint_shape[1:]  # For multitask learning
#             test_shape = torch.Size([joint_shape[0] - self.prediction_strategy.train_shape[0], *tasks_shape])
#
#             # Make the prediction
#             with settings._use_eval_tolerance():
#                 predictive_mean, predictive_covar = self.prediction_strategy.exact_prediction(full_mean, full_covar)
#
#             # Reshape predictive mean to match the appropriate event shape
#             predictive_mean = predictive_mean.view(*batch_shape, *test_shape).contiguous()
#             return full_output.__class__(predictive_mean, predictive_covar)
#
# # default multtask model (with margina likelihood, not scaled task_cov and no coupling
# class MultitaskGPModel(ExactGP):
#     def __init__(self, train_x, train_y, likelihood, n_tasks, rank):
#         super(MultitaskGPModel, self).__init__(train_x, train_y, likelihood)
#
#         # self.mean_module = gpytorch.means.MultitaskMean(          # optional: learn mean function instead of centering
#         #     gpytorch.means.ConstantMean(), num_tasks=n_tasks
#         # )
#         self.mean_module = gpytorch.means.MultitaskMean(
#             gpytorch.means.ZeroMean(), num_tasks=n_tasks
#         )
#
#         self.covar_module = MultitaskKernel(
#             gpytorch.kernels.RBFKernel(), num_tasks=n_tasks, rank=rank
#         )
#
#     def forward(self, x):
#         mean_x = self.mean_module(x)
#         covar_x = self.covar_module(x)
#         return gpytorch.distributions.MultitaskMultivariateNormal(mean_x, covar_x)
#
# class MyMultitaskGPModel(myExactGP):
#     def __init__(self, train_x, train_y, likelihood, n_tasks, rank,
#                  var_constraint = None, covar_factor_constraint = None, lengthscale_constraint = None):
#         super(MyMultitaskGPModel, self).__init__(train_x, train_y, likelihood)
#
#         # self.mean_module = gpytorch.means.MultitaskMean(          # optional: learn mean function instead of centering
#         #     gpytorch.means.ConstantMean(), num_tasks=n_tasks
#         # )
#         self.mean_module = gpytorch.means.MultitaskMean(
#             gpytorch.means.ZeroMean(), num_tasks=n_tasks
#         )
#
#         self.covar_module = myMultitaskKernel(
#             gpytorch.kernels.RBFKernel(lengthscale_constraint = lengthscale_constraint), num_tasks=n_tasks, rank=rank,
#             var_constraint = var_constraint, covar_factor_constraint = covar_factor_constraint
#         )
#
#     def forward(self, x):
#         mean_x = self.mean_module(x)
#         covar_x = self.covar_module(x)
#         return gpytorch.distributions.MultitaskMultivariateNormal(mean_x, covar_x)
#
#
# #    Based on MultitaskKernel: Using a modified IndexKernel - no other changes
# class myMultitaskKernel(Kernel):
#     r"""
#     Kernel supporting Kronecker style multitask Gaussian processes (where every data point is evaluated at every
#     task) using :class:`gpytorch.kernels.IndexKernel` as a basic multitask kernel.
#
#     Given a base covariance module to be used for the data, :math:`K_{XX}`, this kernel computes a task kernel of
#     specified size :math:`K_{TT}` and returns :math:`K = K_{TT} \otimes K_{XX}`. as an
#     :obj:`gpytorch.lazy.KroneckerProductLazyTensor`.
#
#     :param ~gpytorch.kernels.Kernel data_covar_module: Kernel to use as the data kernel.
#     :param int num_tasks: Number of tasks
#     :param int rank: (default 1) Rank of index kernel to use for task covariance matrix.
#     :param ~gpytorch.priors.Prior task_covar_prior: (default None) Prior to use for task kernel.
#         See :class:`gpytorch.kernels.IndexKernel` for details.
#     :param dict kwargs: Additional arguments to pass to the kernel.
#     """
#
#     def __init__(self, data_covar_module, num_tasks, rank=1, task_covar_prior=None,
#                  var_constraint = None, covar_factor_constraint = None, **kwargs):
#         """
#         """
#         super(myMultitaskKernel, self).__init__(**kwargs)
#         self.task_covar_module = myIndexKernel(
#             num_tasks=num_tasks, batch_shape=self.batch_shape, rank=rank, prior=task_covar_prior,
#             var_constraint=var_constraint, covar_factor_constraint = covar_factor_constraint
#         )
#         self.data_covar_module = data_covar_module
#         self.num_tasks = num_tasks
#
#     def forward(self, x1, x2, diag=False, last_dim_is_batch=False, **params):
#         if last_dim_is_batch:
#             raise RuntimeError("MultitaskKernel does not accept the last_dim_is_batch argument.")
#         covar_i = self.task_covar_module.covar_matrix
#         if len(x1.shape[:-2]):
#             covar_i = covar_i.repeat(*x1.shape[:-2], 1, 1)
#         covar_x = lazify(self.data_covar_module.forward(x1, x2, **params))
#         res = KroneckerProductLazyTensor(covar_x, covar_i)
#         return renp.diag() if diag else res
#
#     def num_outputs_per_input(self, x1, x2):
#         """
#         Given `n` data points `x1` and `m` datapoints `x2`, this multitask
#         kernel returns an `(n*num_tasks) x (m*num_tasks)` covariance matrix.
#         """
#         return self.num_tasks
#
# # implement an index kernel B^TB +sigma I
# # based on IndexKernel with changes:
# # - a constant diagonal offset
# # - scaled to correlation matrix
# class myIndexKernel(Kernel):
#     r"""
#     A kernel for discrete indices. Kernel is defined by a lookup table.
#
#     .. math::
#
#         \begin{equation}
#             k(i, j) = \left(BB^\top + \text{diag}(\mathbf v) \right)_{i, j}
#         \end{equation}
#
#     where :math:`B` is a low-rank matrix, and :math:`\mathbf v` is a  non-negative vector.
#     These parameters are learned.
#
#     Args:
#         :attr:`num_tasks` (int):
#             Total number of indices.
#         :attr:`batch_shape` (torch.Size, optional):
#             Set if the MultitaskKernel is operating on batches of data (and you want different
#             parameters for each batch)
#         :attr:`rank` (int):
#             Rank of :math:`B` matrix.
#         :attr:`prior` (:obj:`gpytorch.priors.Prior`):
#             Prior for :math:`B` matrix.
#         :attr:`var_constraint` (Constraint, optional):
#             Constraint for added diagonal component. Default: `Positive`.
#
#     Attributes:
#         covar_factor:
#             The :math:`B` matrix.
#         lov_var:
#             The element-wise log of the :math:`\mathbf v` vector.
#     """
#
#     def __init__(self, num_tasks, rank=1, prior=None, var_constraint=None, covar_factor_constraint = None, **kwargs):
#         if rank > num_tasks:
#             raise RuntimeError("Cannot create a task covariance matrix larger than the number of tasks")
#         super().__init__(**kwargs)
#         self.num_tasks = num_tasks
#
#         if var_constraint is None:
#             var_constraint = Positive()
#
#         self.register_parameter(
#             name="raw_covar_factor", parameter=torch.nn.Parameter(torch.randn(*self.batch_shape, num_tasks, rank))
#         )
#         self.register_parameter(name="raw_var",
#                                 parameter=torch.nn.Parameter(torch.randn(*self.batch_shape, 1)))  # 1 instead of num_tasks
#         if prior is not None:
#             self.register_prior("IndexKernelPrior", prior, self._eval_covar_matrix)
#
#         self.register_constraint("raw_var", var_constraint)
#         if covar_factor_constraint is not None:
#             self.register_constraint("raw_covar_factor", covar_factor_constraint)
#             self.has_covar_factor_constraint = True
#         else:
#             self.has_covar_factor_constraint = False
#
#     @property
#     def var(self):
#         return self.raw_var_constraint.transform(self.raw_var)
#
#     @var.setter
#     def var(self, value):
#         self._set_var(value)
#
#     def _set_var(self, value):
#         self.initialize(raw_var=self.raw_var_constraint.inverse_transform(value))
#
#     @property
#     def covar_factor(self):
#         if self.has_covar_factor_constraint:
#             return self.raw_covar_factor_constraint.transform(self.raw_covar_factor)
#         else:
#             return self.raw_covar_factor
#
#     @covar_factor.setter
#     def covar_factor(self, value):
#         self._set_covar_factor(value)
#
#     def _set_covar_factor(self, value):
#         self.initialize(raw_covar_factor=self.raw_covar_factor_constraint.inverse_transform(value))
#
#
#     def _eval_covar_matrix(self):
#         cf = self.covar_factor
#         C = cf @ cf.transpose(-1, -2) + self.var *  torch.eye(self.num_tasks)   # instead of diag(var)
#         Cdiag = torch.diag(C)
#         C = torch.diag(1 / Cdiag.sqrt()) @ C @ torch.diag(1 / Cdiag.sqrt())       # scale to correlation matrix
#         return C
#
#     @property
#     def covar_matrix(self):
#         # var = self.var
#         # res = PsdSumLazyTensor(RootLazyTensor(self.covar_factor), DiagLazyTensor(var))
#         # return res
#         return self._eval_covar_matrix()
#
#     def forward(self, i1, i2, **params):
#         covar_matrix = self._eval_covar_matrix()
#         batch_shape = _mul_broadcast_shape(i1.shape[:-2], self.batch_shape)
#         index_shape = batch_shape + i1.shape[-2:]
#
#         res = InterpolatedLazyTensor(
#             base_lazy_tensor=covar_matrix,
#             left_interp_indices=i1.expand(index_shape),
#             right_interp_indices=i2.expand(index_shape),
#         )
#         return res
#
#
#
# class ELBO(MarginalLogLikelihood):
#     """
#     Class to implment ELBO as loss function - based on ExactMarginalLogLikelihood
#     """
#
#     def __init__(self, likelihood, model, ZCov):
#         if not isinstance(likelihood, _GaussianLikelihoodBase) and not isinstance(likelihood, _myGaussianLikelihoodBase):
#             raise RuntimeError("Likelihood must be Gaussian for exact inference")
#         super(ELBO, self).__init__(likelihood, model)
#         self.ZCov = ZCov
#
#     def forward(self, function_dist, target, *params):
#         r"""
#         Computes the MLL given :math:`p(\mathbf f)` and :math:`\mathbf y`.
#
#         :param ~gpytorch.distributions.MultivariateNormal function_dist: :math:`p(\mathbf f)`
#             the outputs of the latent function (the :obj:`gpytorch.models.myExactGP`)
#         :param torch.Tensor target: :math:`\mathbf y` The target values
#         :rtype: torch.Tensor
#         :return: Exact MLL. Output shape corresponds to batch shape of the model/input data.
#         """
#         if not isinstance(function_dist, MultivariateNormal):
#             raise RuntimeError("ExactMarginalLogLikelihood can only operate on Gaussian random variables")
#
#         # Get the log prob of the marginal distribution
#         output = self.likelihood(function_dist, *params) # this is N(mu, K + noise_var I) should be N(mu, (1-noise_var) * K + noise_var I) TODO multiply kernel by (1- self.likelihood.raw_noise)
#
#         res = output.log_prob(target)
#         # modify to get N(mu, (1-noise_var) * K + noise_var I)
#
#         # Add additional terms (SGPR / learned inducing points, heteroskedastic likelihood models)
#         for added_loss_term in self.model.added_loss_terms():
#             res = res.add(added_loss_term.loss(*params))
#
#         # Add uncertainty in Z
#         res.add(-0.5 * torch.trace(output.precision_matrix @ self.ZCov))
#
#         # Add log probs of priors on the (functions of) parameters
#         for _, prior, closure, _ in self.named_priors():
#             res.add_(prior.log_prob(closure()).sum())
#
#         # Scale by the amount of data we have
#         num_data = target.size(-1)
#         return res.div_(num_data)
#
#
#
# # Modify liklelihoods: is N(mu, K + noise_var I) should be N(mu, (1-noise_var) * K + noise_var I
# class _myGaussianLikelihoodBase(Likelihood):
#     """Base class for Gaussian Likelihoods, supporting general heteroskedastic noise models."""
#
#     def __init__(self, noise_covar: Noise, **kwargs: Any) -> None:
#
#         super().__init__()
#         param_transform = kwargs.get("param_transform")
#         if param_transform is not None:
#             warnings.warn(
#                 "The 'param_transform' argument is now deprecated. If you want to use a different "
#                 "transformaton, specify a different 'noise_constraint' instead.",
#                 DeprecationWarning,
#             )
#
#         self.noise_covar = noise_covar
#
#     def _shaped_noise_covar(self, base_shape: torch.Size, *params: Any, **kwargs: Any):
#         return self.noise_covar(*params, shape=base_shape, **kwargs)
#
#     def expected_log_prob(self, target: Tensor, input: MultivariateNormal, *params: Any, **kwargs: Any) -> Tensor:
#         mean, variance = input.mean, input.variance
#         num_event_dim = len(input.event_shape)
#
#         noise = self._shaped_noise_covar(mean.shape, *params, **kwargs).diag()
#         # Potentially reshape the noise to deal with the multitask case
#         noise = noise.view(*noise.shape[:-1], *input.event_shape)
#
#         res = ((target - mean) ** 2 + variance) / noise + noise.log() + math.log(2 * math.pi)
#         res = res.mul(-0.5)
#         if num_event_dim > 1:  # Do appropriate summation for multitask Gaussian likelihoods
#             res = res.sum(list(range(-1, -num_event_dim, -1)))
#         return res
#
#     def forward(self, function_samples: Tensor, *params: Any, **kwargs: Any) -> base_distributions.Normal:
#         noise = self._shaped_noise_covar(function_samples.shape, *params, **kwargs).diag()
#         return base_distributions.Normal(function_samples, noise.sqrt())
#
#     def log_marginal(
#         self, observations: Tensor, function_dist: MultivariateNormal, *params: Any, **kwargs: Any
#     ) -> Tensor:
#         marginal = self.marginal(function_dist, *params, **kwargs)
#         # We're making everything conditionally independent
#         indep_dist = base_distributions.Normal(marginal.mean, marginal.variance.clamp_min(1e-8).sqrt())
#         res = indep_dist.log_prob(observations)
#
#         # Do appropriate summation for multitask Gaussian likelihoods
#         num_event_dim = len(function_dist.event_shape)
#         if num_event_dim > 1:
#             res = res.sum(list(range(-1, -num_event_dim, -1)))
#         return res
#
#     def marginal(self, function_dist: MultivariateNormal, *params: Any, **kwargs: Any) -> MultivariateNormal:
#         mean, covar = function_dist.mean, function_dist.lazy_covariance_matrix
#         noise_covar = self._shaped_noise_covar(mean.shape, *params, **kwargs)
#         # print(self.raw_noise)
#         # print(self.noise)
#         # print(self.raw_noise_constraint)
#         full_covar = (1-self.noise) * covar + noise_covar
#         # full_covar = covar + noise_covar
#         return function_dist.__class__(mean, full_covar)
#
# # as rank =0 we do not model noise correlation - most things that happen here are irrelevant
# class _myMultitaskGaussianLikelihoodBase(_myGaussianLikelihoodBase):
#     """Base class for multi-task Gaussian Likelihoods, supporting general heteroskedastic noise models. """
#
#     def __init__(self, num_tasks, noise_covar, rank=0, task_correlation_prior=None, batch_shape=torch.Size()):
#         """
#         Args:
#             num_tasks (int):
#                 Number of tasks.
#             noise_covar (:obj:`gpytorch.module.Module`):
#                 A model for the noise covariance. This can be a simple homoskedastic noise model, or a GP
#                 that is to be fitted on the observed measurement errors.
#             rank (int):
#                 The rank of the task noise covariance matrix to fit. If `rank` is set to 0, then a diagonal covariance
#                 matrix is fit.
#             task_correlation_prior (:obj:`gpytorch.priors.Prior`):
#                 Prior to use over the task noise correlation matrix. Only used when `rank` > 0.
#             batch_shape (torch.Size):
#                 Number of batches.
#         """
#         super().__init__(noise_covar=noise_covar)
#         if rank != 0:
#             if rank > num_tasks:
#                 raise ValueError(f"Cannot have rank ({rank}) greater than num_tasks ({num_tasks})")
#             tidcs = torch.tril_indices(num_tasks, rank, dtype=torch.long)
#             self.tidcs = tidcs[:, 1:]  # (1, 1) must be 1.0, no need to parameterize this
#             task_noise_corr = torch.randn(*batch_shape, self.tidcs.size(-1))
#             self.register_parameter("task_noise_corr", torch.nn.Parameter(task_noise_corr))
#             if task_correlation_prior is not None:
#                 self.register_prior(
#                     "MultitaskErrorCorrelationPrior", task_correlation_prior, lambda: self._eval_corr_matrix
#                 )
#         elif task_correlation_prior is not None:
#             raise ValueError("Can only specify task_correlation_prior if rank>0")
#         self.num_tasks = num_tasks
#         self.rank = rank
#         # # Handle deprecation of parameterization - TODO: Remove in future release
#         # self._register_load_state_dict_pre_hook(deprecate_task_noise_corr)
#
#     def _eval_corr_matrix(self):
#         tnc = self.task_noise_corr
#         fac_diag = torch.ones(*tnc.shape[:-1], self.num_tasks, device=tnc.device, dtype=tnc.dtype)
#         Cfac = torch.diag_embed(fac_diag)
#         Cfac[..., self.tidcs[0], self.tidcs[1]] = self.task_noise_corr
#         # squared rows must sum to one for this to be a correlation matrix
#         C = Cfac / Cfac.pow(2).sum(dim=-1, keepdim=True).sqrt()
#         return C @ C.transpose(-1, -2)
#
#     def _shaped_noise_covar(self, base_shape, *params):
#         if len(base_shape) >= 2:
#             *batch_shape, n, _ = base_shape
#         else:
#             *batch_shape, n = base_shape
#
#         # compute the noise covariance
#         if len(params) > 0:
#             shape = None
#         else:
#             shape = base_shape if len(base_shape) == 1 else base_shape[:-1]
#         noise_covar = self.noise_covar(*params, shape=shape)
#
#         if self.rank > 0:
#             # if rank > 0, compute the task correlation matrix
#             # TODO: This is inefficient, change repeat so it can repeat LazyTensors w/ multiple batch dimensions
#             task_corr = self._eval_corr_matrix()
#             exp_shape = torch.Size([*batch_shape, n]) + task_corr.shape[-2:]
#             task_corr_exp = lazify(task_corr.unsqueeze(-3).expand(exp_shape))
#             noise_sem = noise_covar.sqrt()
#             task_covar_blocks = MatmulLazyTensor(MatmulLazyTensor(noise_sem, task_corr_exp), noise_sem)
#         else:
#             # otherwise tasks are uncorrelated
#             task_covar_blocks = noise_covar
#
#         if len(batch_shape) == 1:
#             # TODO: Properly support general batch shapes in BlockDiagLazyTensor (no shape arithmetic)
#             tcb_eval = task_covar_blocks.evaluate()
#             task_covar = BlockDiagLazyTensor(lazify(tcb_eval), block_dim=-3)
#         else:
#             task_covar = BlockDiagLazyTensor(task_covar_blocks)
#
#         return task_covar
#
#     def forward(self, function_samples: Tensor, *params: Any, **kwargs: Any) -> base_distributions.Normal:
#         noise = self._shaped_noise_covar(function_samples.shape, *params, **kwargs).diag()
#         noise = noise.view(*noise.shape[:-1], *function_samples.shape[-2:])
#         return base_distributions.Independent(base_distributions.Normal(function_samples, noise.sqrt()), 1)
#
#
# class myMultitaskGaussianLikelihood(_myMultitaskGaussianLikelihoodBase):
#     """
#     A convenient extension of the :class:`gpytorch.likelihoods.GaussianLikelihood` to the multitask setting that allows
#     for a full cross-task covariance structure for the noise. The fitted covariance matrix has rank `rank`.
#     If a strictly diagonal task noise covariance matrix is desired, then rank=0 should be set. (This option still
#     allows for a different `log_noise` parameter for each task.). This likelihood assumes homoskedastic noise.
#
#     Like the Gaussian likelihood, this object can be used with exact inference.
#     """
#
#     def __init__(
#         self,
#         num_tasks,
#         rank=0,
#         task_correlation_prior=None,
#         batch_shape=torch.Size(),
#         noise_prior=None,
#         noise_constraint=None,
#     ):
#         """
#         Args:
#             num_tasks (int): Number of tasks.
#
#             rank (int): The rank of the task noise covariance matrix to fit. If `rank` is set to 0,
#             then a diagonal covariance matrix is fit.
#
#             task_correlation_prior (:obj:`gpytorch.priors.Prior`): Prior to use over the task noise correlaton matrix.
#             Only used when `rank` > 0.
#
#         """
#         if noise_constraint is None:
#             noise_constraint = GreaterThan(1e-4)
#
#         noise_covar = MultitaskHomoskedasticNoise(
#             num_tasks=num_tasks, noise_prior=noise_prior, noise_constraint=noise_constraint, batch_shape=batch_shape
#         )
#         super().__init__(
#             num_tasks=num_tasks,
#             noise_covar=noise_covar,
#             rank=rank,
#             task_correlation_prior=task_correlation_prior,
#             batch_shape=batch_shape,
#         )
#
#         self.register_parameter(name="raw_noise", parameter=torch.nn.Parameter(torch.zeros(*batch_shape, 1)))
#         self.register_constraint("raw_noise", noise_constraint)
#
#     @property
#     def noise(self):
#         return self.raw_noise_constraint.transform(self.raw_noise)
#
#     @noise.setter
#     def noise(self, value):
#         self._set_noise(value)
#
#     def _set_noise(self, value):
#         if not torch.is_tensor(value):
#             value = torch.as_tensor(value).to(self.raw_noise)
#         self.initialize(raw_noise=self.raw_noise_constraint.inverse_transform(value))
#
#     def _shaped_noise_covar(self, base_shape, *params):
#         noise_covar = super()._shaped_noise_covar(base_shape, *params)
#         noise = self.noise
#         return noise_covar.add_diag(noise)


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import anndata
import os
import torch
import scipy.sparse as sp
from sklearn.preprocessing import normalize
from tqdm import tqdm

torch_install=False

def global_imports(modulename,shortname = None, asfunction = False):
    if shortname is None: 
        shortname = modulename
    if asfunction is False:
        globals()[shortname] = __import__(modulename)
    else:        
        globals()[shortname] = __import__(modulename)


class scnocd(object):
    def check_torch(self):
        """
        
        """
        global torch_install
        try:
            import torch
            import torch.nn as nn
            import torch.nn.functional as F
            torch_install=True
            print('torch have been install version:',torch.__version__)
        except ImportError:
            raise ImportError(
                'Please install the pytorch: `conda install -c conda-forge pytorch` or `pip install pytorch`.'
            )

    def __init__(self,adata,gpu=0):
        self.check_torch()
        global torch_install
        if torch_install==True:
            global_imports("torch")
            global_imports("torch.nn","nn")
            global_imports("torch.nn.functional","F")
            globals()['nocd'] = __import__("omicverse.nocd",fromlist=['nocd'])
        self.adata_raw=adata
        self.adata=adata.copy()
        self.device  = torch.device(f"cuda:{gpu}") if gpu >= 0 and torch.cuda.is_available() else torch.device('cpu')

      
    def matrix_transform(self,clustertype='leiden'):
        
        try:
            self.adata.obsp['connectivities']
        except NameError:
            var_exists = False
            print('......You need to calculate the neighbor by sc.pp.neighbors')
            return None
        else:
            var_exists = True
        
        try:
            self.adata.obs[clustertype]
        except NameError:
            var_exists = False
            print('......You need to calculate the leiden by sc.tl.leiden or other clustertype')
            return None
        else:
            var_exists = True
            
        self.X=sp.csr_matrix(self.adata.X)
        self.A=self.adata.obsp['connectivities']
        self.clustertype=clustertype
        self.Z_gt=pd.get_dummies(self.adata.obs[clustertype]).values
        self.Z_gt=self.Z_gt.astype(np.float32)

        
        self.N, self.K = self.Z_gt.shape



    def matrix_normalize(self,cuda=False):
        if torch.cuda.is_available():
            cuda=True
        self.x_norm = normalize(self.X)
        self.x_norm=nocd.utils.to_sparse_tensor(self.x_norm,cuda=cuda)

     
    def GNN_configure(self,hidden_size=128,
                     weight_decay=1e-2,
                     dropout=0.5,
                     batch_norm=True,
                     lr=1e-3,
                     max_epochs=500,
                     display_step=25,
                     balance_loss=True,
                     stochastic_loss=True,
                     batch_size=20000):
        
        self.hidden_sizes = [hidden_size]    # hidden sizes of the GNN
        self.weight_decay = weight_decay     # strength of L2 regularization on GNN weights
        self.dropout = dropout           # whether to use dropout
        self.batch_norm = batch_norm       # whether to use batch norm
        self.lr = lr              # learning rate
        self.max_epochs = max_epochs        # number of epochs to train
        self.display_step = display_step       # how often to compute validation loss
        self.balance_loss = balance_loss     # whether to use balanced loss
        self.stochastic_loss = stochastic_loss  # whether to use stochastic or full-batch training
        self.batch_size = batch_size      # batch size (only for stochastic training)

    
    def GNN_preprocess(self,num_workers=5,
                      ):
        self.sampler = nocd.sampler.get_edge_sampler(self.A, self.batch_size, self.batch_size, num_workers=num_workers)
        self.gnn = nocd.nn.GCN(self.x_norm.shape[1], self.hidden_sizes, self.K, 
                               batch_norm=self.batch_norm, dropout=self.dropout).to(self.device)
        self.adj_norm = self.gnn.normalize_adj(self.A).to(self.device)
        self.decoder = nocd.nn.BerpoDecoder(self.N, self.A.nnz, balance_loss=self.balance_loss)
        self.opt = torch.optim.Adam(self.gnn.parameters(), lr=self.lr)

       
    def get_nmi(self,thresh=0.5):
        """Compute Overlapping NMI of the communities predicted by the GNN."""
        self.gnn.eval()
        Z = F.relu(self.gnn(self.x_norm, self.adj_norm))
        Z_pred = Z.cpu().detach().numpy() > thresh
        nmi = nocd.metrics.overlapping_nmi(Z_pred, self.Z_gt)
        return nmi
    
    def save(self,gnn_save_dir:str='save_model',
            gnn_save_name:str='gnn',):
        """
        Saves the trained GNN model.

        Arguments:
            gnn_save_dir: the directory to save the trained GNN model. Default is 'save_model'.
            gnn_save_name: the name of the saved GNN model. Default is 'gnn'.

        """
        path_save = os.path.join(gnn_save_dir, f"{gnn_save_name}.pth")
        if not os.path.exists(gnn_save_dir):
            os.makedirs(gnn_save_dir)
        torch.save(self.gnn.state_dict(), path_save)
        print(f"...save trained gnn in {path_save}.")

    def load(self,gnn_load_dir):
        """
        Loads the trained GNN model.

        Arguments:
            gnn_load_dir: the directory to load the trained GNN model.

        """
        #path_load = os.path.join(gnn_load_dir, "gnn.pth")
        #self.model_saver.load_state_dict(torch.load(path_load))
        print(f'loading model from {gnn_load_dir}')
        self.gnn.load_state_dict(torch.load(gnn_load_dir, map_location=self.device))

    
    
    def GNN_model(self):
        val_loss = np.inf
        validation_fn = lambda: val_loss
        early_stopping = nocd.train.NoImprovementStopping(validation_fn, patience=10)
        self.model_saver = nocd.train.ModelSaver(self.gnn)
        with tqdm(total=self.max_epochs) as t:
            for epoch, batch in enumerate(self.sampler):
                if epoch > self.max_epochs:
                    break
                if epoch % 25 == 0:
                    with torch.no_grad():
                        self.gnn.eval()
                        # Compute validation loss
                        Z = F.relu(self.gnn(self.x_norm, self.adj_norm))
                        val_loss = self.decoder.loss_full(Z, self.A)
                        #print(f'Epoch {epoch:4d}, loss.full = {val_loss:.4f}, nmi = {self.get_nmi():.2f}')
                        t.set_description(f'Epoch {epoch:4d}, loss.full = {val_loss:.4f}, nmi = {self.get_nmi():.2f}')

                        # Check if it's time for early stopping / to save the model
                        early_stopping.next_step()
                        if early_stopping.should_save():
                            self.model_saver.save()
                        if early_stopping.should_stop():
                            print(f'Breaking due to early stopping at epoch {epoch}')
                            break

                # Training step
                self.gnn.train()
                self.opt.zero_grad()
                Z = F.relu(self.gnn(self.x_norm, self.adj_norm))
                ones_idx, zeros_idx = batch
                if self.stochastic_loss:
                    loss = self.decoder.loss_batch(Z, ones_idx, zeros_idx)
                else:
                    loss = self.decoder.loss_full(Z, A)
                loss += nocd.utils.l2_reg_loss(self.gnn, scale=self.weight_decay)
                loss.backward()
                self.opt.step()

           
    def GNN_result(self,thresh=0.5):
        thresh = thresh

        Z = F.relu(self.gnn(self.x_norm, self.adj_norm))
        self.Z_pred = Z.cpu().detach().numpy() > thresh
        self.model_saver.restore()
        print(f'Final nmi = {self.get_nmi(thresh):.3f}')

      
    def GNN_plot(self,figsize=[10,10],markersize=0.05):
        plt.figure(figsize=figsize)
        z = np.argmax(self.Z_pred, 1)
        o = np.argsort(z)
        nocd.utils.plot_sparse_clustered_adjacency(self.A, self.K, z, o, markersize=markersize)

       
       
    def calculate_nocd(self):
        zpred=self.Z_pred+0
        pr=np.argmax(zpred, axis=-1)
        prf=pd.DataFrame(pr)
        prf.index=self.adata.obs.index

        pred=pd.DataFrame(zpred)
        pred.index=self.adata.obs.index
        m=(pred==1).sum(axis=1)
        n=pd.DataFrame(m)
        n[n>1]=-1
        
        df=n.merge(prf, how='inner', left_index=True, right_index=True)
        df1=df.sort_values("0_x")
        
        k=n.loc[~(n==1).all(axis=1)]
        df1.drop(df1.head(len(k)).index,inplace=True) 
        df2=df1.drop(["0_x"],axis=1)
        df3=df2.rename(columns={"0_y":0})
        
        con=pd.concat([k,df3],axis=0)
        
        print('......add nocd result to adata.obs')
        self.adata.obs['nocd']=con
    
    
    def cal_nocd(self):
        #pred matrix
        pred_pd=pd.DataFrame(self.Z_pred+0)
        pred_pd.index=self.adata.obs.index

        #nocd result
        nocd_res=[]
        for cell in pred_pd.index:
            nocd_type=''
            for i in pred_pd.loc[cell][pred_pd.loc[cell]!=0].index:
                if nocd_type=='':
                    nocd_type+=str(i)
                else:
                    nocd_type+='-'+str(i)
            nocd_res.append(nocd_type)
        nocd_res
        print('......add nocd result to adata.obs')
        self.adata.obs['nocd_n']=nocd_res
        self.adata=self.adata[self.adata.obs['nocd_n']!='']

    def get_pair_dict(self,):
        self.cal_nocd()
        return dict(zip(self.adata[~self.adata.obs['nocd_n'].str.contains('-')].obs[[self.clustertype]].values.reshape(-1),
        self.adata[~self.adata.obs['nocd_n'].str.contains('-')].obs[['nocd_n']].values.reshape(-1)))
    
        



from ._preprocess import (identify_robust_genes,
                          select_hvf_pegasus,
                          highly_variable_features,
                          remove_cc_genes,
                          preprocess,
                          normalize_pearson_residuals,
                          highly_variable_genes,
                          scale,
                          regress,
                          regress_and_scale,
                          neighbors,
                          pca)

from ._qc import quantity_control,qc,filter_cells,filter_genes


r"""
single (A omic framework for single cell omic analysis)
"""

from ._cosg import cosg
from ._anno import pySCSA,MetaTiME,scanpy_lazy,scanpy_cellanno_from_dict,get_celltype_marker
from ._nocd import scnocd
from ._mofa import pyMOFAART,pyMOFA,GLUE_pair,factor_exact,factor_correlation,get_weights,glue_pair
from ._scdrug import autoResolution,writeGEP,Drug_Response
from ._cpdb import cpdb_network_cal,cpdb_plot_network,cpdb_plot_interaction,cpdb_interaction_filtered,cpdb_submeans_exacted
from ._scgsea import geneset_aucell,pathway_aucell,pathway_aucell_enrichment,pathway_enrichment,pathway_enrichment_plot
from ._via import pyVIA,scRNA_hematopoiesis
from ._simba import pySIMBA
from ._tosica import pyTOSICA
from ._atac import atac_concat_get_index,atac_concat_inner,atac_concat_outer
from ._batch import batch_correction
from ._cellfategenie import cellfategenie,gene_trends
from ._ltnn import scLTNN,plot_origin_tesmination,find_related_gene
from ._traj import TrajInfer


import pandas as pd
import torch
import numpy as np
import anndata
import matplotlib.pyplot as plt
from ._utils import load_data, data_process,bulk2single_data_prepare
from ._vae import train_vae, generate_vae, load_vae
from ._map_utils import create_data, DFRunner, joint_analysis, knn,create_data_pyomic
import os
import warnings
import matplotlib
from typing import Union,Tuple
warnings.filterwarnings("ignore")

class Single2Spatial(object):

    def __init__(self,single_data:anndata.AnnData,
                 spatial_data:anndata.AnnData,
                 celltype_key:str,
                 spot_key:list=['xcoord','ycoord'],
                 top_marker_num=500,
                marker_used=True,gpu:Union[int,str]=0) -> None:
        """Init Single2Spatial model

        Arguments:
            single_data: the anndata object of single cell data
            spatial_data: the anndata object of spatial data 
            celltype_key: the key of cell type in `single_data.obs`
            spot_key: the key of spot in `spatial_data.obs` Default: ['xcoord','ycoord']
            top_marker_num: the number of top marker genes used in the model. Default: 500
            marker_used: whether use marker genes in the model. Default: True
            gpu: the gpu used in the model. Default: 0 if gpu is available, else cpu, mps is also supported
        """

        self.single_data = single_data
        self.spatial_data = spatial_data
        self.top_marker_num = top_marker_num
        self.marker_used = marker_used
        self.celltype_key=celltype_key
        if gpu=='mps' and torch.backends.mps.is_available():
            print('Note that mps may loss will be nan, used it when torch is supported')
            self.used_device = torch.device("mps")
        else:
            self.used_device = torch.device(f"cuda:{gpu}") if gpu >= 0 and torch.cuda.is_available() else torch.device('cpu')
        self.history=[]
        self.input_data = create_data_pyomic(self.single_data,self.spatial_data,
                                             celltype_key,spot_key,)

    def train(self,spot_num:int,
                    cell_num:int,
                    df_save_dir:str='save_model',
                    df_save_name:str='df',
                    max_cell_in_diff_spot_ratio=None,
                    k:int=10,
                    random_seed:int=112,
                    mul_train:int=1,save=True,
                    n_jobs:int=1,num_epochs=1000,batch_size=1000,predicted_size=32)->anndata.AnnData:
        """Train the model of single2spatial

        Arguments:
            spot_num: the number of spots in the spatial data predicted
            cell_num: the number of cells in each spot predicted
            df_save_dir: the directory to save the model
            df_save_name: the name of the model
            max_cell_in_diff_spot_ratio: the ratio of max cell number in different spot
            k: the number of nearest neighbors
            random_seed: the random seed
            mul_train: the number of times to train the model
            n_jobs: the number of jobs to run in parallel

        Returns:
            sp_adata: the anndata object of the predicted spatial data
        
        """
        # load data

        xtrain, ytrain = create_data(self.input_data['input_sc_meta'], 
                                     self.input_data['input_sc_data'], self.input_data["input_st_data"], 
                                     spot_num, cell_num,
                                     self.top_marker_num,
                                     self.marker_used, mul_train)
        df_runner = DFRunner(self.input_data['input_sc_data'], self.input_data['input_sc_meta'], 
                             self.input_data['input_st_data'], self.input_data['input_st_meta'], 
                             self.marker_used, self.top_marker_num, random_seed=random_seed,n_jobs=n_jobs,device=self.used_device)
        self.df_runner=df_runner
        df_meta, df_spot = self.df_runner.run(xtrain, ytrain, max_cell_in_diff_spot_ratio, k, df_save_dir, 
                                              df_save_name,num_epochs=num_epochs,batch_size=batch_size,predicted_size=predicted_size)
        
        if save:
            path_save = os.path.join(df_save_dir, f"{df_save_name}.pth")
            if not os.path.exists(df_save_dir):
                os.makedirs(df_save_dir)
            torch.save(df_runner.model.state_dict(), path_save)
            print(f"...save trained net in {path_save}.")

        sp_adata=anndata.AnnData(df_spot.T)
        sp_adata.obs=df_meta
        sp_adata.obs.set_index(sp_adata.obs['Cell'],inplace=True)
        sp_adata.obsm['X_spatial']=sp_adata.obs[['Cell_xcoord','Cell_ycoord']].values
        self.sp_adata=sp_adata
        return sp_adata
        #  save df
        os.makedirs(map_save_dir, exist_ok=True)
        meta_dir = os.path.join(map_save_dir, f'meta_{map_save_name}_{k}.csv')
        spot_dir = os.path.join(map_save_dir, f'data_{map_save_name}_{k}.csv')
        df_meta.to_csv(meta_dir)
        df_spot.to_csv(spot_dir)
        print(f"saving result to {meta_dir} and {spot_dir}")
        return df_meta, df_spot


        return df_meta, df_spot
    
    def save(self,df_save_dir:str='save_model',
                df_save_name:str='df',):
        """Save the model of single2spatial
        
        Arguments:
            df_save_dir: the directory to save the model
            df_save_name: the name of the model
        """

        path_save = os.path.join(df_save_dir, f"{df_save_name}.pth")
        if not os.path.exists(df_save_dir):
            os.makedirs(df_save_dir)
        torch.save(self.df_runner.model.state_dict(), path_save)
        print(f"...save trained net in {path_save}.")
        #print("Model have been saved to "+os.path.join(df_save_dir, f"{df_save_name}"))

    
    def load(self,modelsize,
                    df_load_dir:str='save_model/df',
                    
                    max_cell_in_diff_spot_ratio=None,
                    k:int=10,
                    random_seed:int=112,
                    n_jobs:int=1,predicted_size=32)->anndata.AnnData:
        """Load the model of single2spatial

        Arguments:
            spot_num: the number of spots in the spatial data predicted
            cell_num: the number of cells in each spot predicted
            df_load_dir: the directory to load the model
            max_cell_in_diff_spot_ratio: the ratio of max cell number in different spot
            k: the number of nearest neighbors
            random_seed: the random seed
            mul_train: the number of times to train the model

        Returns:
            sp_adata: the anndata object of the predicted spatial data
        
        """
        #xtrain, ytrain = create_data(self.input_data['input_sc_meta'], 
        #                             self.input_data['input_sc_data'], self.input_data["input_st_data"], 
        #                             spot_num, cell_num,
        #                             self.top_marker_num,
        #                             self.marker_used, mul_train)
        df_runner = DFRunner(self.input_data['input_sc_data'], self.input_data['input_sc_meta'], 
                             self.input_data['input_st_data'], self.input_data['input_st_meta'], 
                             self.marker_used, self.top_marker_num, random_seed=random_seed,n_jobs=n_jobs)
        self.df_runner=df_runner
        df_meta, df_spot = self.df_runner.load(df_load_dir,modelsize,max_cell_in_diff_spot_ratio, k,
                                              predicted_size=predicted_size)

        sp_adata=anndata.AnnData(df_spot.T)
        sp_adata.obs=df_meta
        sp_adata.obs.set_index(sp_adata.obs['Cell'],inplace=True)
        sp_adata.obsm['X_spatial']=sp_adata.obs[['Cell_xcoord','Cell_ycoord']].values
        self.sp_adata=sp_adata
        return sp_adata
        #  save df
        os.makedirs(map_save_dir, exist_ok=True)
        meta_dir = os.path.join(map_save_dir, f'meta_{map_save_name}_{k}.csv')
        spot_dir = os.path.join(map_save_dir, f'data_{map_save_name}_{k}.csv')
        df_meta.to_csv(meta_dir)
        df_spot.to_csv(spot_dir)
        print(f"saving result to {meta_dir} and {spot_dir}")
        return df_meta, df_spot
    
    def spot_assess(self)->anndata.AnnData:
        """Assess the predicted spatial data
        
        Returns:
            sp_adata_spot: the anndata object of the predicted spatial data with spot-level information
        """

        # spot-level
        # calculate cell type proportion per spot
        prop = self.sp_adata.obs[['Cell', 'Cell_type', 'Spot']].pivot_table(index=['Spot'], columns=['Cell_type'], aggfunc='count',values = 'Cell', fill_value=0)
        prop = prop.div(prop.sum(axis=1), axis=0)
        prop.columns = pd.Index(list(prop.columns))
        prop['Spot_xcoord'] = np.array(pd.pivot_table(self.sp_adata.obs,values='Spot_xcoord',index=['Spot'])['Spot_xcoord'].values)
        prop['Spot_ycoord'] = np.array(pd.pivot_table(self.sp_adata.obs,values='Spot_ycoord',index=['Spot'])['Spot_ycoord'].values)

        # aggregate gene expression per spot
        pred_spot_new = self.sp_adata.to_df()
        genes = pred_spot_new.columns
        pred_spot_new['Spot'] = self.sp_adata.obs['Spot']
        pred_spot_mean = pred_spot_new.groupby('Spot')[genes].mean()

        sp_adata_spot=anndata.AnnData(pred_spot_mean)
        sp_adata_spot.obs=prop
        sp_adata_spot.obsm['X_spatial']=sp_adata_spot.obs[['Spot_xcoord','Spot_ycoord']].values
        return sp_adata_spot

def spatial_mapping(generate_sc_meta,
                    generate_sc_data,
                    input_st_data_path,
                    input_st_meta_path,
                    map_save_dir='output',  # file_dir
                    map_save_name='map',  # file_name
                    ):
        input_st_data = pd.read_csv(input_st_data_path, index_col=0)
        input_st_meta = pd.read_csv(input_st_meta_path, index_col=0)
        print('start to process image-based st data...')
        sc_gene_new = generate_sc_data._stat_axis.values.tolist()
        st_gene_new = input_st_data._stat_axis.values.tolist()
        intersect_gene_new = list(set(sc_gene_new).intersection(set(st_gene_new)))
        generate_sc_data_new = generate_sc_data.loc[intersect_gene_new]
        input_st_data_new = input_st_data.loc[intersect_gene_new]

        sc_cell_rename = [f'SC_{i}' for i in range(1, generate_sc_data_new.shape[1] + 1)]
        generate_sc_data.columns = generate_sc_data_new.columns = sc_cell_rename
        generate_sc_meta = generate_sc_meta.drop(['Cell'], axis=1)
        generate_sc_meta.insert(0, 'Cell', sc_cell_rename)
        generate_sc_meta['Batch'] = 'sc'
        generate_sc_meta_new = generate_sc_meta.drop(['Cell_type'], axis=1)
        st_cell_rename = [f'ST_{i}' for i in range(1, input_st_data_new.shape[1] + 1)]
        input_st_data.columns = input_st_data_new.columns = st_cell_rename
        input_st_meta = input_st_meta.drop(['Cell'], axis=1)
        input_st_meta.insert(0, 'Cell', st_cell_rename)
        input_st_meta_new = pd.DataFrame({'Cell': st_cell_rename, 'Batch': 'st'})

        all_data = generate_sc_data_new.join(input_st_data_new)
        all_meta = pd.concat([generate_sc_meta_new, input_st_meta_new], ignore_index=True)
        joint_data = joint_analysis(all_data, all_meta['Batch'], ref_batch="st")
        joint_data[joint_data < 0] = 0
        sc_data_new = joint_data.iloc[:, 0:generate_sc_data_new.shape[1]]
        st_data_new = joint_data.iloc[:, generate_sc_data_new.shape[1]:all_data.shape[1]]

        _, ind = knn(data=sc_data_new.T, query=st_data_new.T, k=10)

        st_data_pred = pd.DataFrame()
        st_meta_pred = pd.DataFrame(columns=['Cell', 'Cell_type'])

        for i in range(len(st_cell_rename)):
            st_data_pred[st_cell_rename[i]] = list(generate_sc_data.iloc[:, ind[i]].mean(axis=1))
            ct_tmp = list(generate_sc_meta.iloc[ind[i], :].Cell_type)
            ct_pred = max(ct_tmp, key=ct_tmp.count)
            st_meta_pred.loc[st_cell_rename[i]] = [st_cell_rename[i], ct_pred]

        st_data_pred.index = generate_sc_data.index
        st_meta_pred = pd.merge(st_meta_pred, input_st_meta, how='left', on='Cell')

        #  save df
        os.makedirs(map_save_dir, exist_ok=True)
        meta_dir = os.path.join(map_save_dir, f'meta_{map_save_name}.csv')
        data_dir = os.path.join(map_save_dir, f'data_{map_save_name}.csv')
        st_meta_pred.to_csv(os.path.join(meta_dir))
        st_data_pred.to_csv(os.path.join(data_dir))
        print(f'saving to {meta_dir} and {data_dir}')
        return st_meta_pred, st_data_pred



from ..via.core import VIA
from ..via.datasets_via import scRNA_hematopoiesis
import anndata
import numpy as np
import scanpy as sc
import pandas as pd

import igraph as ig
import pygam as pg
from datetime import datetime
from typing import Union,Tuple

from ..via.utils_via import *
from ..via.plotting_via import *
#from ..via.utils_via import sc_loc_ofsuperCluster_PCAspace,compute_velocity_on_grid

def hematopoiesis()->anndata.AnnData:
    return scRNA_hematopoiesis()

class pyVIA(object):

    def __init__(self,adata:anndata.AnnData,adata_key:str='X_pca',adata_ncomps:int=80,basis:str='X_umap',
                 clusters:str='',dist_std_local:float=2, jac_std_global=0.15, labels:np.ndarray=None,
                 keep_all_local_dist='auto', too_big_factor:float=0.4, resolution_parameter:float=1.0, partition_type:str="ModularityVP", small_pop:int=10,
                 jac_weighted_edges:bool=True, knn:int=30, n_iter_leiden:int=5, random_seed:int=42,
                 num_threads=-1, distance='l2', time_smallpop=15,
                 super_cluster_labels:bool=False,                 super_node_degree_list:bool=False, super_terminal_cells:bool=False, x_lazy:float=0.95, alpha_teleport:float=0.99,
                 root_user=None, preserve_disconnected:bool=True, dataset:str='', super_terminal_clusters:list=[],
                 is_coarse=True, csr_full_graph:np.ndarray='', csr_array_locally_pruned='', ig_full_graph='',
                 full_neighbor_array='', full_distance_array='',  df_annot=None,
                 preserve_disconnected_after_pruning:bool=False,
                 secondary_annotations:list=None, pseudotime_threshold_TS:int=30, cluster_graph_pruning_std:float=0.15,
                 visual_cluster_graph_pruning:float=0.15, neighboring_terminal_states_threshold=3, num_mcmc_simulations=1300,
                 piegraph_arrow_head_width=0.1,
                 piegraph_edgeweight_scalingfactor=1.5, max_visual_outgoing_edges:int=2, via_coarse=None, velocity_matrix=None,
                 gene_matrix=None, velo_weight=0.5, edgebundle_pruning=None, A_velo = None, CSM = None, edgebundle_pruning_twice=False, pca_loadings = None, time_series=False,
                 time_series_labels:list=None, knn_sequential:int = 10, knn_sequential_reverse:int = 0,t_diff_step:int = 1,single_cell_transition_matrix = None,
                 embedding_type:str='via-mds',do_compute_embedding:bool=False, color_dict:dict=None,user_defined_terminal_cell:list=[], user_defined_terminal_group:list=[],
                 do_gaussian_kernel_edgeweights:bool=False,RW2_mode:bool=False,working_dir_fp:str ='/home/shobi/Trajectory/Datasets/') -> None:
        r"""
        Initialize a pyVIA object.

        Arguments:
            adata: An AnnData object containing the scRNA-seq.
            adata_key: the key of the AnnData in obsm to perform VIA on. default: 'X_pca'
            adata_ncomps: the number of components to use from the AnnData in obsm to perform VIA on. default: 80
            basis: the key of the AnnData in obsm to use as the basis for the embedding. default: 'X_umap'
            clusters: the clusters to use for the VIA analysis. default: ''
            dist_std_local: local level of pruning for PARC graph clustering stage. Range (0.1,3) higher numbers mean more edge retention
            jac_std_global: (optional, default = 0.15, can also set as 'median') global level graph pruning for PARC clustering stage. Number of standard deviations below the network’s mean-jaccard-weighted edges. 0.1-1 provide reasonable pruning.higher value means less pruning (more edges retained). e.g. a value of 0.15 means all edges that are above mean(edgeweight)-0.15*std(edge-weights) are retained. We find both 0.15 and ‘median’ to yield good results/starting point and resulting in pruning away ~ 50-60% edges
            labels: default is None. and PARC clusters are used for the viagraph. alternatively provide a list of clustermemberships that are integer values (not strings) to construct the viagraph using another clustering method or available annotations
            keep_all_local_dist: default value of 'auto' means that for smaller datasets local-pruning is done prior to clustering, but for large datasets local pruning is set to False for speed.
            
            too_big_factor: (optional, default=0.4). Forces clusters > 0.4*n_cells to be re-clustered
            resolution_parameter: (float) the resolution parameter for the Louvain algorithm.
            partition_type: (str, default: "ModularityVP") the partitioning algorithm to use.
            small_pop: (int, default: 10) the number of cells to be considered in a small population.
            jac_weighted_edges: (bool, default: True) whether to use weighted edges in the PARC clustering step.
            knn: (int, optional, default: 30) the number of K-Nearest Neighbors for HNSWlib KNN graph. Larger knn means more graph connectivity. Lower knn means more loosely connected clusters/cells.
            n_iter_leiden: (int) the number of iterations for the Leiden algorithm.
            random_seed: (int) the random seed to pass to the clustering algorithm.
            num_threads: (int) the number of threads to use for the clustering algorithm.
            distance: (str, default: 'l2') the distance metric to use for graph construction and similarity. Options are 'l2', 'ip', and 'cosine'.
            visual_cluster_graph_pruning: (float, default: 0.15) the pruning level for the cluster graph. This only comes into play if the user deliberately chooses not to use the default edge-bundling method of visualizating edges (draw_piechart_graph()) and instead calls draw_piechart_graph_nobundle(). It controls the number of edges plotted for visual effect. This does not impact computation of terminal states, pseudotime, or lineage likelihoods.
            cluster_graph_pruning_std: (float, default: 0.15) the pruning level of the cluster graph. Often set to the same value as the PARC clustering level of jac_std_global. Reasonable range is [0.1, 1]. To retain more connectivity in the clustergraph underlying the trajectory computations, increase the value.
            time_smallpop: (max time to be allowed handling singletons) the maximum time allowed to handle singletons.
            super_cluster_labels:
            super_node_degree_list:
            super_terminal_cells:
            x_lazy: (float, default: 0.95) 1-x = probability of staying in the same node (lazy). Values between 0.9-0.99 are reasonable.
            alpha_teleport: (float, default: 0.99) 1-alpha is probability of jumping. Values between 0.95-0.99 are reasonable unless prior knowledge of teleportation.
            root_user: (list, None) the root user list. Can be a list of strings, a list of int, or None. The default is None. When the root_user is set as None and an RNA velocity matrix is available, a root will be automatically computed. If the root_user is None and no velocity matrix is provided, then an arbitrary root is selected. If the root_user is ['celltype_earlystage'] where the str corresponds to an item in true_label, then a suitable starting point will be selected corresponding to this group.
            preserve_disconnected: bool (default = True) If you believe there may be disconnected trajectories then set this to False
            dataset: str Can be set to 'group' or '' (default). this refers to the type of root label (group level root or single cell index) you are going to provide. if your true_label has a sensible group of cells for a root then you can set dataset to 'group' and make the root parameter ['labelname_root_cell_type'] if your root corresponds to one particular cell then set dataset = '' (default)
            embedding: ndarray (optional, default = None) embedding (e.g. precomputed tsne, umap, phate, via-umap) for plotting data. Size n_cells x 2 If an embedding is provided when running VIA, then a scatterplot colored by pseudotime, highlighting terminal fates
            velo_weight: float (optional, default = 0.5) #float between [0,1]. the weight assigned to directionality and connectivity derived from scRNA-velocity
            neighboring_terminal_states_threshold:int (default = 3). Candidates for terminal states that are neighbors of each other may be removed from the list if they have this number of more of terminal states as neighbors
            knn_sequential:int (default =10) number of knn in the adjacent time-point for time-series data (t_i and t_i+1)
            knn_sequential_reverse: int (default = 0) number of knn enforced from current to previous time point
            t_diff_step: int (default =1) Number of permitted temporal intervals between connected nodes. If time data is labeled as [0,25,50,75,100,..] then t_diff_step=1 corresponds to '25' and only edges within t_diff_steps are retained
            is_coarse:bool (default = True) If running VIA in two iterations where you wish to link the second fine-grained iteration with the initial iteration, then you set to False
            via_coarse: VIA (default = None) If instantiating a second iteration of VIA that needs to be linked to a previous iteration (e.g. via0), then set via_coarse to the previous via0 object
            df_annot: DataFrame (default None) used for the Mouse Organ data
            preserve_disconnected_after_pruning:bool (default = True) If you believe there are disconnected trajectories then set this to False and test your hypothesis
            A_velo: ndarray Cluster Graph Transition matrix based on rna velocity [n_clus x n_clus]
            velocity_matrix: matrix (default None) matrix of size [n_samples x n_genes]. this is the velocity matrix computed by scVelo (or similar package) and stored in adata.layers['velocity']. The genes used for computing velocity should correspond to those useing in gene_matrix Requires gene_matrix to be provided too.
            gene_matrix: matrix (default None) Only used if Velocity_matrix is available. matrix of size [n_samples x n_genes]. We recommend using a subset like HVGs rather than full set of genes. (need to densify input if taking from adata = adata.X.todense())
            time_series: if the data has time-series labels then set to True
            time_series_labels:list (default None) list of integer values of temporal annotations corresponding to e.g. hours (post fert), days, or sequential ordering
            pca_loadings: array (default None) the loadings of the pcs used to project the cells (to projected euclidean location based on velocity). n_cells x n_pcs
            secondary_annotations: None (default None)
            edgebundle_pruning:float (default=None) will by default be set to the same as the cluster_graph_pruning_std and influences the visualized level of pruning of edges. Typical values can be between [0,1] with higher numbers retaining more edges
            edgebundle_pruning_twice:bool default: False. When True, the edgebundling is applied to a further visually pruned (visual_cluster_graph_pruning) and can sometimes simplify the visualization. it does not impact the pseudotime and lineage computations
            piegraph_arrow_head_width: float (default = 0.1) size of arrow heads in via cluster graph
            piegraph_edgeweight_scalingfactor: (defaulf = 1.5) scaling factor for edge thickness in via cluster graph
            max_visual_outgoing_edges: int (default =2) Rarely comes into play. Only used if the user later chooses to plot the via-graph without edgebunding using draw_piechart_graph_nobundle() Only allows max_visual_outgoing_edges to come out of any given node.
            edgebundle_pruning:float (default=None) will by default be set to the same as the cluster_graph_pruning_std and influences the visualized level of pruning of edges. Typical values can be between [0,1] with higher numbers retaining more edges
            edgebundle_pruning_twice:bool default: False. When True, the edgebundling is applied to a further visually pruned (visual_cluster_graph_pruning) and can sometimes simplify the visualization. it does not impact the pseudotime and lineage computations
            pseudotime_threshold_TS: int (default = 30) corresponds to the criteria for a state to be considered a candidate terminal cell fate to be 30% or later of the computed pseudotime range
            num_mcmc_simulations:int (default = 1300) number of random walk simulations conducted
            embedding_type: str (default = 'via-mds', other options are 'via-umap' and 'via-force'
            do_compute_embedding: bool (default = False) If you want an embedding (n_samples x2) to be computed on the basis of the via sc graph then set this to True
            do_gaussian_kernel_edgeweights: bool (default = False) Type of edgeweighting on the graph edges

        
        """


        self.adata = adata
        #self.adata_key = adata_key
        data = adata.obsm[adata_key][:, 0:adata_ncomps]
        embedding=self.adata.obsm[basis]
        true_label=adata.obs[clusters]
        self.clusters=clusters
        self.basis=basis
        
        if root_user is not None:
             dataset='group'

        

        self.model=VIA(data=data,true_label=true_label,
                 dist_std_local=dist_std_local,jac_std_global=jac_std_global,labels=labels,
                 keep_all_local_dist=keep_all_local_dist,too_big_factor=too_big_factor,resolution_parameter=resolution_parameter,partition_type=partition_type,small_pop=small_pop,
                 jac_weighted_edges=jac_weighted_edges,knn=knn,n_iter_leiden=n_iter_leiden,random_seed=random_seed,
                 num_threads=num_threads,distance=distance,time_smallpop=time_smallpop,
                 super_cluster_labels=super_cluster_labels,super_node_degree_list=super_node_degree_list,super_terminal_cells=super_terminal_cells,x_lazy=x_lazy,alpha_teleport=alpha_teleport,
                 root_user=root_user,preserve_disconnected=preserve_disconnected,dataset=dataset,super_terminal_clusters=super_terminal_clusters,
                 is_coarse=is_coarse,csr_full_graph=csr_full_graph,csr_array_locally_pruned=csr_array_locally_pruned,ig_full_graph=ig_full_graph,
                 full_neighbor_array=full_neighbor_array,full_distance_array=full_distance_array,embedding=embedding,df_annot=df_annot,
                 preserve_disconnected_after_pruning=preserve_disconnected_after_pruning,
                 secondary_annotations=secondary_annotations,pseudotime_threshold_TS=pseudotime_threshold_TS,cluster_graph_pruning_std=cluster_graph_pruning_std,
                 visual_cluster_graph_pruning=visual_cluster_graph_pruning,neighboring_terminal_states_threshold=neighboring_terminal_states_threshold,num_mcmc_simulations=num_mcmc_simulations,
                 piegraph_arrow_head_width=piegraph_arrow_head_width,
                 piegraph_edgeweight_scalingfactor=piegraph_edgeweight_scalingfactor,max_visual_outgoing_edges=max_visual_outgoing_edges,via_coarse=via_coarse,velocity_matrix=velocity_matrix,
                 gene_matrix=gene_matrix,velo_weight=velo_weight,edgebundle_pruning=edgebundle_pruning,A_velo=A_velo,CSM=CSM,edgebundle_pruning_twice=edgebundle_pruning_twice,pca_loadings=pca_loadings,time_series=time_series,
                 time_series_labels=time_series_labels,knn_sequential=knn_sequential,knn_sequential_reverse=knn_sequential_reverse,t_diff_step=t_diff_step,single_cell_transition_matrix=single_cell_transition_matrix,
                 embedding_type=embedding_type,do_compute_embedding=do_compute_embedding,color_dict=color_dict,user_defined_terminal_cell=user_defined_terminal_cell,user_defined_terminal_group=user_defined_terminal_group,
                 do_gaussian_kernel_edgeweights=do_gaussian_kernel_edgeweights,RW2_mode=RW2_mode,working_dir_fp=working_dir_fp
                 )
    def run(self):
        """calculate the via graph and pseudotime
        
        """
        self.model.run_VIA()

    def get_piechart_dict(self,label:int=0,clusters:str='')->dict:
        """
        Cluster composition graph

        Arguments:
            label: int (default=0) cluster label of pie chart
            clusters: the celltype you want interested
        
        Returns:
            res_dict: cluster composition graph
        """
        if clusters=='':
            clusters=self.clusters
        self.adata.obs[clusters]=self.adata.obs[clusters].astype('category')
        cluster_i_loc=np.where(np.asarray(self.model.labels) == label)[0]
        res_dict=dict(self.adata.obs.iloc[cluster_i_loc].value_counts(clusters))
        return res_dict
    
    def get_pseudotime(self,adata=None):
        """
        Extract the pseudotime of VIA

        Arguments:
            adata: an adata object of you interested,if None, it will be added to `self.adata.obs['pt_via']`

        """

        print('...the pseudotime of VIA added to AnnData obs named `pt_via`')
        if adata is None:
            self.adata.obs['pt_via']=self.model.single_cell_pt_markov
        else:
            adata.obs['pt_via']=self.model.single_cell_pt_markov

    def plot_piechart_graph(self,clusters:str='', type_data='pt',
                                gene_exp:list=[], title='', 
                                cmap:str=None, ax_text=True, figsize:tuple=(8,4),
                                dpi=150,headwidth_arrow = 0.1, 
                                alpha_edge=0.4, linewidth_edge=2, 
                                edge_color='darkblue',reference=None, 
                                show_legend:bool=True, pie_size_scale:float=0.8, fontsize:float=8)->Tuple[matplotlib.figure.Figure,
                                                                                                          matplotlib.axes._axes.Axes,
                                                                                                          matplotlib.axes._axes.Axes]:
        """plot two subplots with a clustergraph level representation of the viagraph showing true-label composition (lhs) and pseudotime/gene expression (rhs)
    
        Arguments:
            clusters : column name of the adata.obs dataframe that contains the cluster labels
            type_data : string  default 'pt' for pseudotime colored nodes. or 'gene'
            gene_exp : list of values (column of dataframe) corresponding to feature or gene expression to be used to color nodes at CLUSTER level
            title : string
            cmap : default None. automatically chooses coolwarm for gene expression or viridis_r for pseudotime
            ax_text : Bool default= True. Annotates each node with cluster number and population of membership
            dpi : int default = 150
            headwidth_bundle : default = 0.1. width of arrowhead used to directed edges
            reference : None or list. list of categorical (str) labels for cluster composition of the piecharts (LHS subplot) length = n_samples.
            pie_size_scale : float default=0.8 scaling factor of the piechart nodes
            fontsize : float default=8. fontsize of the text in the piecharts
            figsize : tuple default=(8,4). size of the figure

        Returns:
            fig: Returns matplotlib figure with two axes that plot the clustergraph using edge bundling
            ax: left axis shows the clustergraph with each node colored by annotated ground truth membership.
            ax1: right axis shows the same clustergraph with each node colored by the pseudotime or gene expression
        """


        if clusters=='':
            clusters=self.clusters
        self.adata.obs[clusters]=self.adata.obs[clusters].astype('category')
        fig, ax, ax1 = draw_piechart_graph_pyomic(clusters=clusters,adata=self.adata,
                                   via_object=self.model, type_data=type_data,
                                gene_exp=gene_exp, title=title, 
                                cmap=cmap, ax_text=ax_text,figsize=figsize,
                                dpi=dpi,headwidth_arrow = headwidth_arrow,
                                alpha_edge=alpha_edge, linewidth_edge=linewidth_edge,
                                edge_color=edge_color,reference=reference,
                                show_legend=show_legend, pie_size_scale=pie_size_scale, fontsize=fontsize)
        return fig, ax, ax1
    
    def plot_stream(self,clusters:str='',basis:str='',
                   density_grid:float=0.5, arrow_size:float=0.7, arrow_color:str = 'k',
                   arrow_style="-|>",  max_length:int=4, linewidth:float=1,min_mass = 1, cutoff_perc:int = 5,
                   scatter_size:int=500, scatter_alpha:float=0.5,marker_edgewidth:float=0.1,
                   density_stream:int = 2, smooth_transition:int=1, smooth_grid:float=0.5,
                   color_scheme:str = 'annotation', add_outline_clusters:bool=False,
                   cluster_outline_edgewidth = 0.001,gp_color = 'white', bg_color='black' ,
                   dpi=80 , title='Streamplot', b_bias=20, n_neighbors_velocity_grid=None,
                   other_labels:list = None,use_sequentially_augmented:bool=False, cmap_str:str='rainbow')->Tuple[matplotlib.figure.Figure,
                                                                                                          matplotlib.axes._axes.Axes]:
        """Construct vector streamplot on the embedding to show a fine-grained view of inferred directions in the trajectory

        Arguments:
            clusters : column name of the adata.obs dataframe that contains the cluster labels
            basis : str, default = 'X_umap', which to use for the embedding
            density_grid : float, default = 0.5, density of the grid on which to project the directionality of cells
            arrow_size : float, default = 0.7, size of the arrows in the streamplot
            arrow_color : str, default = 'k', color of the arrows in the streamplot
            arrow_style : str, default = "-|>", style of the arrows in the streamplot
            max_length : int, default = 4, maximum length of the arrows in the streamplot
            linewidth : float, default = 1, width of  lines in streamplot
            min_mass : float, default = 1, minimum mass of the arrows in the streamplot
            cutoff_perc : int, default = 5, cutoff percentage of the arrows in the streamplot
            scatter_size : int, default = 500, size of scatter points
            scatter_alpha : float, default = 0.5, transpsarency of scatter points
            marker_edgewidth : float, default = 0.1, width of outline arround each scatter point
            density_stream : int, default = 2, density of the streamplot
            smooth_transition : int, default = 1, smoothness of the transition between the streamplot and the scatter points
            smooth_grid : float, default = 0.5, smoothness of the grid on which to project the directionality of cells
            color_scheme : str, default = 'annotation' corresponds to self.true_labels. Other options are 'time' (uses single-cell pseudotime) or 'clusters' (uses self.clusters)
            add_outline_clusters : bool, default = False, whether to add an outline to the clusters
            cluster_outline_edgewidth : float, default = 0.001, width of the outline around the clusters
            gp_color : str, default = 'white', color of the grid points
            bg_color : str, default = 'black', color of the background
            dpi : int, default = 80, dpi of the figure
            title : str, default = 'Streamplot', title of the figure
            b_bias : int, default = 20, higher value makes the forward bias of pseudotime stronger
            n_neighbors_velocity_grid : int, default = None, number of neighbors to use for the velocity grid
            other_labels : list, default = None, list of other labels to plot in the streamplot
            use_sequentially_augmented : bool, default = False, whether to use the sequentially augmented data
            cmap_str : str, default = 'rainbow', color map to use for the streamplot

        Returns:
            fig : matplotlib figure
            ax : matplotlib axis
        """

        if clusters=='':
            clusters=self.clusters
        if basis=='':
            basis=self.basis
        self.adata.obs[clusters]=self.adata.obs[clusters].astype('category')
        embedding=self.adata.obsm[basis]
        fig,ax = via_streamplot_pyomic(adata=self.adata,clusters=clusters,via_object=self.model, 
                                 embedding=embedding,density_grid=density_grid, arrow_size=arrow_size,
                                 arrow_color=arrow_color,arrow_style=arrow_style,  max_length=max_length,
                                 linewidth=linewidth,min_mass = min_mass, cutoff_perc=cutoff_perc,
                                 scatter_size=scatter_size, scatter_alpha=scatter_alpha,marker_edgewidth=marker_edgewidth,
                                 density_stream=density_stream, smooth_transition=smooth_transition, smooth_grid=smooth_grid,
                                 color_scheme=color_scheme, add_outline_clusters=add_outline_clusters,
                                 cluster_outline_edgewidth = cluster_outline_edgewidth,gp_color = gp_color, bg_color=bg_color,
                                 dpi=dpi , title=title, b_bias=b_bias, n_neighbors_velocity_grid=n_neighbors_velocity_grid,
                                 other_labels=other_labels,use_sequentially_augmented=use_sequentially_augmented, cmap_str=cmap_str)
        return fig,ax

    def plot_trajectory_gams(self,clusters:str='',basis:str='',via_fine=None, idx=None,
                         title_str:str= "Pseudotime", draw_all_curves:bool=True, arrow_width_scale_factor:float=15.0,
                         scatter_size:float=50, scatter_alpha:float=0.5,figsize:tuple=(8,4),
                         linewidth:float=1.5, marker_edgewidth:float=1, cmap_pseudotime:str='viridis_r',dpi:int=80,
                         highlight_terminal_states:bool=True, use_maxout_edgelist:bool =False)->Tuple[matplotlib.figure.Figure,
                                                                                                 matplotlib.axes._axes.Axes,
                                                                                                 matplotlib.axes._axes.Axes]:
        """projects the graph based coarse trajectory onto a umap/tsne embedding

        Arguments:
            clusters : column name of the adata.obs dataframe that contains the cluster labels
            basis : str, default = 'X_umap', which to use for the embedding
            via_fine : via object suggest to use via_object only unless you found that running via_fine gave better pathways
            idx : default: None. Or List. if you had previously computed a umap/tsne (embedding) only on a subset of the total n_samples (subsampled as per idx), then the via objects and results will be indexed according to idx too
            title_str : title of figure
            draw_all_curves : if the clustergraph has too many edges to project in a visually interpretable way, set this to False to get a simplified view of the graph pathways
            arrow_width_scale_factor : the width of the arrows is proportional to the edge weight. This factor scales the width of the arrows
            scatter_size : size of the scatter points
            scatter_alpha : transparency of the scatter points
            linewidth : width of the lines
            marker_edgewidth : width of the outline around each scatter point
            cmap_pseudotime : color map to use for the pseudotime
            dpi : dpi of the figure
            highlight_terminal_states :  whether or not to highlight/distinguish the clusters which are detected as the terminal states by via

        Returns:
            fig : matplotlib figure
            ax1 : matplotlib axis
            ax2 : matplotlib axis

        """


        if clusters=='':
            clusters=self.clusters
        if basis=='':
            basis=self.basis
        self.adata.obs[clusters]=self.adata.obs[clusters].astype('category')
        embedding=self.adata.obsm[basis]
        fig,ax1,ax2 = draw_trajectory_gams_pyomic(adata=self.adata,clusters=clusters,via_object=self.model, 
                                            via_fine=via_fine, embedding=embedding, idx=idx,
                                            title_str=title_str, draw_all_curves=draw_all_curves, arrow_width_scale_factor=arrow_width_scale_factor,
                                            scatter_size=scatter_size, scatter_alpha=scatter_alpha,figsize=figsize,
                                            linewidth=linewidth, marker_edgewidth=marker_edgewidth, cmap_pseudotime=cmap_pseudotime,dpi=dpi,
                                            highlight_terminal_states=highlight_terminal_states, use_maxout_edgelist=use_maxout_edgelist)
        return fig,ax1,ax2
    
    def plot_lineage_probability(self,clusters:str='',basis:str='',via_fine=None, 
                                idx=None, figsize:tuple=(8,4),
                                cmap:str='plasma', dpi:int=80, scatter_size =None,
                                marker_lineages:list = [], fontsize:int=12)->Tuple[matplotlib.figure.Figure,
                                                                                   matplotlib.axes._axes.Axes]:
        """G is the igraph knn (low K) used for shortest path in high dim space. no idx needed as it's made on full sample, knn_hnsw is the knn made in the embedded space used for query to find the nearest point in the downsampled embedding that corresponds to the single cells in the full graph
        
        Arguments:
            clusters : column name of the adata.obs dataframe that contains the cluster labels
            basis : str, default = 'X_umap', which to use for the embedding
            via_fine : usually just set to same as via_coarse unless you ran a refined run and want to link it to initial via_coarse's terminal clusters
            idx : if one uses a downsampled embedding of the original data, then idx is the selected indices of the downsampled samples used in the visualization
            figsize : size of the figure
            cmap : color map to use for the lineage probability
            dpi : dpi of the figure
            scatter_size : size of the scatter points
            marker_lineages : Default is to use all lineage pathways. other provide a list of lineage number (terminal cluster number).
            fontsize : fontsize of the title

        Returns:
            fig : matplotlib figure
            axs : matplotlib axis
        """


        if clusters=='':
            clusters=self.clusters
        if basis=='':
            basis=self.basis
        self.adata.obs[clusters]=self.adata.obs[clusters].astype('category')
        embedding=self.adata.obsm[basis]
        fig, axs = draw_sc_lineage_probability(via_object=self.model,via_fine=via_fine, embedding=embedding,figsize=figsize,
                                               idx=idx, cmap_name=cmap, dpi=dpi, scatter_size =scatter_size,
                                            marker_lineages = marker_lineages, fontsize=fontsize)
        fig.tight_layout()
        return fig, axs
    
    def plot_gene_trend(self,gene_list:list=None,figsize:tuple=(8,4),
                        magic_steps:int=3, spline_order:int=5, dpi:int=80,cmap:str='jet', 
                        marker_genes:list = [], linewidth:float = 2.0,
                        n_splines:int=10,  fontsize_:int=12, marker_lineages=[])->Tuple[matplotlib.figure.Figure,
                                                                                        matplotlib.axes._axes.Axes]:
        """plots the gene expression trend along the pseudotime

        Arguments:
            gene_list : list of genes to plot
            figsize : size of the figure
            magic_steps : number of magic steps to use for imputation
            spline_order : order of the spline to use for smoothing
            dpi : dpi of the figure
            cmap : color map to use for the gene expression
            marker_genes : Default is to use all genes in gene_exp. other provide a list of marker genes that will be used from gene_exp.
            linewidth : width of the lines
            n_splines : number of splines to use for smoothing
            fontsize_ : fontsize of the title
            marker_lineages : Default is to use all lineage pathways. other provide a list of lineage number (terminal cluster number).

        Returns:
            fig : matplotlib figure
            axs : matplotlib axis
        
        """

        df_magic = self.model.do_impute(self.adata[:,gene_list].to_df(), magic_steps=magic_steps, gene_list=gene_list)
        fig, axs=get_gene_expression_pyomic(self.model,df_magic,spline_order=spline_order,dpi=dpi,
                                   cmap=cmap, marker_genes=marker_genes, linewidth=linewidth,figsize=figsize,
                                   n_splines=n_splines,  fontsize_=fontsize_, marker_lineages=marker_lineages)
        fig.tight_layout()
        return fig, axs
    
    def plot_clustergraph(self,gene_list:list,arrow_head:float=0.1,figsize:tuple=(8,4),dpi=80,magic_steps=3,
                          edgeweight_scale:float=1.5, cmap=None, label_=True,)->Tuple[matplotlib.figure.Figure,
                                                                                        matplotlib.axes._axes.Axes]:
        """plot the gene in pie chart for each cluster

        Arguments:
            gene_list : list of genes to plot
            arrow_head : size of the arrow head
            figsize : size of the figure
            edgeweight_scale : scale of the edge weight
            cmap : color map to use for the gene expression
            label_ : whether to label the nodes

        Returns:
            fig : matplotlib figure
            axs : matplotlib axis
        """
        df_magic = self.model.do_impute(self.adata[:,gene_list].to_df(), magic_steps=magic_steps, gene_list=gene_list)
        df_magic['parc'] = self.model.labels
        df_magic_cluster = df_magic.groupby('parc', as_index=True).mean()
        fig, axs = draw_clustergraph_pyomic(via_object=self.model, type_data='gene', gene_exp=df_magic_cluster, 
                                    gene_list=gene_list, arrow_head=arrow_head,figsize=figsize,
                                    edgeweight_scale=edgeweight_scale, cmap=cmap, label_=label_,dpi=dpi)
        fig.tight_layout()
        return fig,axs
    
    def plot_gene_trend_heatmap(self,gene_list:list,marker_lineages:list = [], 
                             fontsize:int=8,cmap:str='viridis', normalize:bool=True, ytick_labelrotation:int = 0, 
                             figsize:tuple=(2,4))->Tuple[matplotlib.figure.Figure,
                                                                    list]:
        """Plot the gene trends on heatmap: a heatmap is generated for each lineage (identified by terminal cluster number). Default selects all lineages

        Arguments:
            gene_list : list of genes to plot
            marker_lineages : list default = None and plots all detected all lineages. Optionally provide a list of integers corresponding to the cluster number of terminal cell fates
            fontsize : int default = 8
            cmap : str default = 'viridis'
            normalize : bool = True
            ytick_labelrotation : int default = 0
            figsize : size of the figure

        Returns:
            fig : matplotlib figure
            axs : list of matplotlib axis       
        """
        
        df_magic = self.model.do_impute(self.adata[:,gene_list].to_df(), magic_steps=3, gene_list=gene_list)
        df_magic['parc'] = self.model.labels
        df_magic_cluster = df_magic.groupby('parc', as_index=True).mean()
        fig,axs=plot_gene_trend_heatmaps_pyomic(via_object=self.model, df_gene_exp=df_magic, 
                                                cmap=cmap,fontsize=fontsize,normalize=normalize,
                                                ytick_labelrotation=ytick_labelrotation,figsize=figsize,
                                                marker_lineages=marker_lineages)
        fig.tight_layout()
        return fig,axs















def via_streamplot_pyomic(adata,clusters,via_object, embedding:np.ndarray=None , density_grid:float=0.5, arrow_size:float=0.7, arrow_color:str = 'k',
arrow_style="-|>",  max_length:int=4, linewidth:float=1,min_mass = 1, cutoff_perc:int = 5,
                   scatter_size:int=500, scatter_alpha:float=0.5,marker_edgewidth:float=0.1, 
                   density_stream:int = 2, smooth_transition:int=1, smooth_grid:float=0.5, 
                   color_scheme:str = 'annotation', add_outline_clusters:bool=False, 
                   cluster_outline_edgewidth = 0.001,gp_color = 'white', bg_color='black' , 
                   dpi=80 , title='Streamplot', b_bias=20, n_neighbors_velocity_grid=None, 
                   other_labels:list = None,use_sequentially_augmented=False, cmap_str:str='rainbow'):
    '''
    Construct vector streamplot on the embedding to show a fine-grained view of inferred directions in the trajectory

    :param via_object:
    :param embedding:  np.ndarray of shape (n_samples, 2) umap or other 2-d embedding on which to project the directionality of cells
    :param density_grid:
    :param arrow_size:
    :param arrow_color:
    :param arrow_style:
    :param max_length:
    :param linewidth:  width of  lines in streamplot, default = 1
    :param min_mass:
    :param cutoff_perc:
    :param scatter_size: size of scatter points default =500
    :param scatter_alpha: transpsarency of scatter points
    :param marker_edgewidth: width of outline arround each scatter point, default = 0.1
    :param density_stream:
    :param smooth_transition:
    :param smooth_grid:
    :param color_scheme: str, default = 'annotation' corresponds to self.true_labels. Other options are 'time' (uses single-cell pseudotime) and 'cluster' (via cluster graph) and 'other'
    :param add_outline_clusters:
    :param cluster_outline_edgewidth:
    :param gp_color:
    :param bg_color:
    :param dpi:
    :param title:
    :param b_bias: default = 20. higher value makes the forward bias of pseudotime stronger
    :param n_neighbors_velocity_grid:
    :param other_labels: list (will be used for the color scheme)
    :param use_sequentially_augmented:
    :param cmap_str:
    :return: fig, ax
    '''
    """
    
   
  
   Parameters
   ----------
   X_emb:

   scatter_size: int, default = 500

   linewidth:

   marker_edgewidth: 

   streamplot matplotlib.pyplot instance of fine-grained trajectories drawn on top of scatter plot
   """

    import matplotlib.patheffects as PathEffects
    if embedding is None:
        embedding = via_object.embedding
        if embedding is None:
            print(f'{datetime.now()}\tWARNING: please assign ambedding attribute to via_object as v0.embedding = ndarray of [n_cells x 2]')

    V_emb = via_object._velocity_embedding(embedding, smooth_transition,b=b_bias, use_sequentially_augmented=use_sequentially_augmented)

    V_emb *=20 #5


    X_grid, V_grid = compute_velocity_on_grid(
        X_emb=embedding,
        V_emb=V_emb,
        density=density_grid,
        smooth=smooth_grid,
        min_mass=min_mass,
        autoscale=False,
        adjust_for_stream=True,
        cutoff_perc=cutoff_perc, n_neighbors=n_neighbors_velocity_grid )

    # adapted from : https://github.com/theislab/scvelo/blob/1805ab4a72d3f34496f0ef246500a159f619d3a2/scvelo/plotting/velocity_embedding_grid.py#L27
    lengths = np.sqrt((V_grid ** 2).sum(0))

    linewidth = 1 if linewidth is None else linewidth
    #linewidth *= 2 * lengths / np.percentile(lengths[~np.isnan(lengths)],90)
    linewidth *= 2 * lengths / lengths[~np.isnan(lengths)].max()

    #linewidth=0.5
    fig, ax = plt.subplots(dpi=dpi)
    ax.grid(False)
    ax.streamplot(X_grid[0], X_grid[1], V_grid[0], V_grid[1], color=arrow_color, arrowsize=arrow_size, arrowstyle=arrow_style, zorder = 3, linewidth=linewidth, density = density_stream, maxlength=max_length)

    #num_cluster = len(set(super_cluster_labels))

    if add_outline_clusters:
        # add black outline to outer cells and a white inner rim
        #adapted from scanpy (scVelo utils adapts this from scanpy)
        gp_size = (2 * (scatter_size * cluster_outline_edgewidth *.1) + 0.1*scatter_size) ** 2

        bg_size = (2 * (scatter_size * cluster_outline_edgewidth)+ math.sqrt(gp_size)) ** 2

        ax.scatter(embedding[:, 0],embedding[:, 1], s=bg_size, marker=".", c=bg_color, zorder=-2)
        ax.scatter(embedding[:, 0], embedding[:, 1], s=gp_size, marker=".", c=gp_color, zorder=-1)

    if color_scheme == 'time':
        ax.scatter(embedding[:,0],embedding[:,1], c=via_object.single_cell_pt_markov,alpha=scatter_alpha,  zorder = 0, s=scatter_size, linewidths=marker_edgewidth, cmap = 'viridis_r')
    else:
        if color_scheme == 'annotation':color_labels = via_object.true_label
        if color_scheme == 'cluster': color_labels= via_object.labels
        if other_labels is not None: color_labels = other_labels

        cmap_ = plt.get_cmap(cmap_str)
        #plt.cm.rainbow(color)

        line = np.linspace(0, 1, len(set(color_labels)))
        
        color_true_list=adata.uns['{}_colors'.format(clusters)]
        line = range(len(color_labels))
        for color, group in zip(line, list(adata.obs[clusters].cat.categories)):
            where = np.where(np.array(color_labels) == group)[0]
            ax.scatter(embedding[where, 0], embedding[where, 1], label=group,
                        c=color_true_list[color],
                        alpha=scatter_alpha,  zorder = 0, s=scatter_size, linewidths=marker_edgewidth)

            x_mean = embedding[where, 0].mean()
            y_mean = embedding[where, 1].mean()
            ax.text(x_mean, y_mean, str(group), fontsize=5, zorder=4, path_effects = [PathEffects.withStroke(linewidth=1, foreground='w')], weight = 'bold')

    fig.patch.set_visible(False)
    ax.axis('off')
    ax.set_title(title)
    return fig, ax


def draw_trajectory_gams_pyomic(adata,clusters,via_object, via_fine=None, embedding: np.ndarray=None, idx=None,
                         title_str:str= "Pseudotime", draw_all_curves:bool=True, arrow_width_scale_factor:float=15.0,
                         scatter_size:float=50, scatter_alpha:float=0.5,figsize:tuple=(8,4),
                         linewidth:float=1.5, marker_edgewidth:float=1, cmap_pseudotime:str='viridis_r',dpi:int=80,highlight_terminal_states:bool=True, use_maxout_edgelist:bool =False):
    '''

    projects the graph based coarse trajectory onto a umap/tsne embedding

    :param via_object: via object
    :param via_fine: via object suggest to use via_object only unless you found that running via_fine gave better pathways
    :param embedding: 2d array [n_samples x 2] with x and y coordinates of all n_samples. Umap, tsne, pca OR use the via computed embedding via_object.embedding
    :param idx: default: None. Or List. if you had previously computed a umap/tsne (embedding) only on a subset of the total n_samples (subsampled as per idx), then the via objects and results will be indexed according to idx too
    :param title_str: title of figure
    :param draw_all_curves: if the clustergraph has too many edges to project in a visually interpretable way, set this to False to get a simplified view of the graph pathways
    :param arrow_width_scale_factor:
    :param scatter_size:
    :param scatter_alpha:
    :param linewidth:
    :param marker_edgewidth:
    :param cmap_pseudotime:
    :param dpi: int default = 150. Use 300 for paper figures
    :param highlight_terminal_states: whether or not to highlight/distinguish the clusters which are detected as the terminal states by via
    :return: f, ax1, ax2
    '''

    if embedding is None:
        embedding = via_object.embedding
        if embedding is None: print(f'{datetime.now()}\t ERROR please provide an embedding or compute using via_mds() or via_umap()')
    if via_fine is None:
        via_fine = via_object
    from mpl_toolkits.axes_grid1 import make_axes_locatable

    if idx is None: idx = np.arange(0, via_object.nsamples)
    cluster_labels = list(np.asarray(via_fine.labels)[idx])
    super_cluster_labels = list(np.asarray(via_object.labels)[idx])
    super_edgelist = via_object.edgelist
    if use_maxout_edgelist==True:
        super_edgelist =via_object.edgelist_maxout
    true_label = list(np.asarray(via_fine.true_label)[idx])
    #true_label=list(adata.obs[clusters].cat.categories)
    knn = via_fine.knn
    ncomp = via_fine.ncomp
    if len(via_fine.revised_super_terminal_clusters)>0:
        final_super_terminal = via_fine.revised_super_terminal_clusters
    else: final_super_terminal = via_fine.terminal_clusters

    sub_terminal_clusters = via_fine.terminal_clusters

    
    sc_pt_markov = list(np.asarray(via_fine.single_cell_pt_markov)[idx])
    super_root = via_object.root[0]



    sc_supercluster_nn = sc_loc_ofsuperCluster_PCAspace(via_object, via_fine, np.arange(0, len(cluster_labels)))
    # draw_all_curves. True draws all the curves in the piegraph, False simplifies the number of edges
    # arrow_width_scale_factor: size of the arrow head
    X_dimred = embedding * 1. / np.max(embedding, axis=0)
    x = X_dimred[:, 0]
    y = X_dimred[:, 1]
    max_x = np.percentile(x, 90)
    noise0 = max_x / 1000

    df = pd.DataFrame({'x': x, 'y': y, 'cluster': cluster_labels, 'super_cluster': super_cluster_labels,
                       'projected_sc_pt': sc_pt_markov},
                      columns=['x', 'y', 'cluster', 'super_cluster', 'projected_sc_pt'])
    df_mean = df.groupby('cluster', as_index=False).mean()
    sub_cluster_isin_supercluster = df_mean[['cluster', 'super_cluster']]

    sub_cluster_isin_supercluster = sub_cluster_isin_supercluster.sort_values(by='cluster')
    sub_cluster_isin_supercluster['int_supercluster'] = sub_cluster_isin_supercluster['super_cluster'].round(0).astype(
        int)

    df_super_mean = df.groupby('super_cluster', as_index=False).mean()
    pt = df_super_mean['projected_sc_pt'].values

    f, (ax1, ax2) = plt.subplots(1, 2, sharey=True, figsize=figsize,dpi=dpi)
    num_true_group = len(set(true_label))
    color_true_list=adata.uns['{}_colors'.format(clusters)]
    
    num_cluster = len(set(super_cluster_labels))
    line = np.linspace(0, 1, num_true_group)
    
    line = range(len(true_label))
    for color, group in zip(line, list(adata.obs[clusters].cat.categories)):
        where = np.where(np.array(true_label) == group)[0]
        ax1.scatter(X_dimred[where, 0], X_dimred[where, 1], label=group, c=color_true_list[color],
                    alpha=scatter_alpha, s=scatter_size, linewidths=marker_edgewidth*.1)  # 10 # 0.5 and 4
    ax1.legend(fontsize=6, frameon = False)
    ax1.set_title('True Labels: ncomps:' + str(ncomp) + '. knn:' + str(knn))
    
    G_orange = ig.Graph(n=num_cluster, edges=super_edgelist)
    ll_ = []  # this can be activated if you intend to simplify the curves
    for fst_i in final_super_terminal:

        path_orange = G_orange.get_shortest_paths(super_root, to=fst_i)[0]
        len_path_orange = len(path_orange)
        for enum_edge, edge_fst in enumerate(path_orange):
            if enum_edge < (len_path_orange - 1):
                ll_.append((edge_fst, path_orange[enum_edge + 1]))

    edges_to_draw = super_edgelist if draw_all_curves else list(set(ll_))
    for e_i, (start, end) in enumerate(edges_to_draw):
        if pt[start] >= pt[end]:
            start, end = end, start

        x_i_start = df[df['super_cluster'] == start]['x'].values
        y_i_start = df[df['super_cluster'] == start]['y'].values
        x_i_end = df[df['super_cluster'] == end]['x'].values
        y_i_end = df[df['super_cluster'] == end]['y'].values


        super_start_x = X_dimred[sc_supercluster_nn[start], 0]
        super_end_x = X_dimred[sc_supercluster_nn[end], 0]
        super_start_y = X_dimred[sc_supercluster_nn[start], 1]
        super_end_y = X_dimred[sc_supercluster_nn[end], 1]
        direction_arrow = -1 if super_start_x > super_end_x else 1

        minx = min(super_start_x, super_end_x)
        maxx = max(super_start_x, super_end_x)

        miny = min(super_start_y, super_end_y)
        maxy = max(super_start_y, super_end_y)

        x_val = np.concatenate([x_i_start, x_i_end])
        y_val = np.concatenate([y_i_start, y_i_end])

        idx_keep = np.where((x_val <= maxx) & (x_val >= minx))[0]
        idy_keep = np.where((y_val <= maxy) & (y_val >= miny))[0]

        idx_keep = np.intersect1d(idy_keep, idx_keep)

        x_val = x_val[idx_keep]
        y_val = y_val[idx_keep]

        super_mid_x = (super_start_x + super_end_x) / 2
        super_mid_y = (super_start_y + super_end_y) / 2
        from scipy.spatial import distance

        very_straight = False
        straight_level = 3
        noise = noise0
        x_super = np.array(
            [super_start_x, super_end_x, super_start_x, super_end_x, super_start_x, super_end_x, super_start_x,
             super_end_x, super_start_x + noise, super_end_x + noise,
             super_start_x - noise, super_end_x - noise])
        y_super = np.array(
            [super_start_y, super_end_y, super_start_y, super_end_y, super_start_y, super_end_y, super_start_y,
             super_end_y, super_start_y + noise, super_end_y + noise,
             super_start_y - noise, super_end_y - noise])

        if abs(minx - maxx) <= 1:
            very_straight = True
            straight_level = 10
            x_super = np.append(x_super, super_mid_x)
            y_super = np.append(y_super, super_mid_y)

        for i in range(straight_level):  # DO THE SAME FOR A MIDPOINT TOO
            y_super = np.concatenate([y_super, y_super])
            x_super = np.concatenate([x_super, x_super])

        list_selected_clus = list(zip(x_val, y_val))

        if len(list_selected_clus) >= 1 & very_straight:
            dist = distance.cdist([(super_mid_x, super_mid_y)], list_selected_clus, 'euclidean')
            k = min(2, len(list_selected_clus))
            midpoint_loc = dist[0].argsort()[:k]

            midpoint_xy = []
            for i in range(k):
                midpoint_xy.append(list_selected_clus[midpoint_loc[i]])

            noise = noise0 * 2

            if k == 1:
                mid_x = np.array([midpoint_xy[0][0], midpoint_xy[0][0] + noise, midpoint_xy[0][0] - noise])
                mid_y = np.array([midpoint_xy[0][1], midpoint_xy[0][1] + noise, midpoint_xy[0][1] - noise])
            if k == 2:
                mid_x = np.array(
                    [midpoint_xy[0][0], midpoint_xy[0][0] + noise, midpoint_xy[0][0] - noise, midpoint_xy[1][0],
                     midpoint_xy[1][0] + noise, midpoint_xy[1][0] - noise])
                mid_y = np.array(
                    [midpoint_xy[0][1], midpoint_xy[0][1] + noise, midpoint_xy[0][1] - noise, midpoint_xy[1][1],
                     midpoint_xy[1][1] + noise, midpoint_xy[1][1] - noise])
            for i in range(3):
                mid_x = np.concatenate([mid_x, mid_x])
                mid_y = np.concatenate([mid_y, mid_y])

            x_super = np.concatenate([x_super, mid_x])
            y_super = np.concatenate([y_super, mid_y])
        x_val = np.concatenate([x_val, x_super])
        y_val = np.concatenate([y_val, y_super])

        x_val = x_val.reshape((len(x_val), -1))
        y_val = y_val.reshape((len(y_val), -1))
        xp = np.linspace(minx, maxx, 500)

        gam50 = pg.LinearGAM(n_splines=4, spline_order=3, lam=10).gridsearch(x_val, y_val)
        XX = gam50.generate_X_grid(term=0, n=500)
        preds = gam50.predict(XX)

        idx_keep = np.where((xp <= (maxx)) & (xp >= (minx)))[0]
        ax2.plot(XX, preds, linewidth=linewidth, c='#323538')  # 3.5#1.5


        mean_temp = np.mean(xp[idx_keep])
        closest_val = xp[idx_keep][0]
        closest_loc = idx_keep[0]

        for i, xp_val in enumerate(xp[idx_keep]):
            if abs(xp_val - mean_temp) < abs(closest_val - mean_temp):
                closest_val = xp_val
                closest_loc = idx_keep[i]
        step = 1

        head_width = noise * arrow_width_scale_factor  # arrow_width needs to be adjusted sometimes # 40#30  ##0.2 #0.05 for mESC #0.00001 (#for 2MORGAN and others) # 0.5#1
        if direction_arrow == 1:
            ax2.arrow(xp[closest_loc], preds[closest_loc], xp[closest_loc + step] - xp[closest_loc],
                      preds[closest_loc + step] - preds[closest_loc], shape='full', lw=0, length_includes_head=False,
                      head_width=head_width, color='#323538')

        else:
            ax2.arrow(xp[closest_loc], preds[closest_loc], xp[closest_loc - step] - xp[closest_loc],
                      preds[closest_loc - step] - preds[closest_loc], shape='full', lw=0, length_includes_head=False,
                      head_width=head_width, color='#323538')

    c_edge = []
    width_edge = []
    pen_color = []
    super_cluster_label = []
    terminal_count_ = 0
    dot_size = []

    for i in sc_supercluster_nn:
        if i in final_super_terminal:
            print(f'{datetime.now()}\tSuper cluster {i} is a super terminal with sub_terminal cluster',
                  sub_terminal_clusters[terminal_count_])
            c_edge.append('yellow')  # ('yellow')
            if highlight_terminal_states == True:
                width_edge.append(2)
                super_cluster_label.append('TS' + str(sub_terminal_clusters[terminal_count_]))
            else:
                width_edge.append(0)
                super_cluster_label.append('')
            pen_color.append('black')
            # super_cluster_label.append('TS' + str(i))  # +'('+str(i)+')')
             # +'('+str(i)+')')
            dot_size.append(60)  # 60
            terminal_count_ = terminal_count_ + 1
        else:
            width_edge.append(0)
            c_edge.append('black')
            pen_color.append('red')
            super_cluster_label.append(str(' '))  # i or ' '
            dot_size.append(20)  # 20

    ax2.set_title(title_str)

    im2 =ax2.scatter(X_dimred[:, 0], X_dimred[:, 1], c=sc_pt_markov, cmap=cmap_pseudotime,  s=0.01)
    divider = make_axes_locatable(ax2)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    f.colorbar(im2, cax=cax, orientation='vertical', label='pseudotime') #to avoid lines drawn on the colorbar we need an image instance without alpha variable
    ax2.scatter(X_dimred[:, 0], X_dimred[:, 1], c=sc_pt_markov, cmap=cmap_pseudotime, alpha=scatter_alpha,
                s=scatter_size, linewidths=marker_edgewidth*.1)
    count_ = 0
    loci = [sc_supercluster_nn[key] for key in sc_supercluster_nn]
    for i, c, w, pc, dsz, lab in zip(loci, c_edge, width_edge, pen_color, dot_size,
                                     super_cluster_label):  # sc_supercluster_nn
        ax2.scatter(X_dimred[i, 0], X_dimred[i, 1], c='black', s=dsz, edgecolors=c, linewidth=w)
        ax2.annotate(str(lab), xy=(X_dimred[i, 0], X_dimred[i, 1]))
        count_ = count_ + 1
    
    ax1.grid(False)
    ax2.grid(False)
    f.patch.set_visible(False)
    ax1.axis('off')
    ax2.axis('off')
    return f, ax1, ax2


def draw_piechart_graph_pyomic(adata,clusters,via_object, type_data='pt',
                                gene_exp:list=[], title='', 
                                cmap:str=None, ax_text=True, figsize=(8,4),
                                dpi=150,headwidth_arrow = 0.1, 
                                alpha_edge=0.4, linewidth_edge=2, 
                                edge_color='darkblue',reference=None, 
                                show_legend:bool=True, pie_size_scale:float=0.8, fontsize:float=8):
    '''
    plot two subplots with a clustergraph level representation of the viagraph showing true-label composition (lhs) and pseudotime/gene expression (rhs)
    Returns matplotlib figure with two axes that plot the clustergraph using edge bundling
    left axis shows the clustergraph with each node colored by annotated ground truth membership.
    right axis shows the same clustergraph with each node colored by the pseudotime or gene expression
    :param via_object: is class VIA (the same function also exists as a method of the class and an external plotting function
    :param type_data: string  default 'pt' for pseudotime colored nodes. or 'gene'
    :param gene_exp: list of values (column of dataframe) corresponding to feature or gene expression to be used to color nodes at CLUSTER level
    :param title: string
    :param cmap: default None. automatically chooses coolwarm for gene expression or viridis_r for pseudotime
    :param ax_text: Bool default= True. Annotates each node with cluster number and population of membership
    :param dpi: int default = 150
    :param headwidth_bundle: default = 0.1. width of arrowhead used to directed edges
    :param reference: None or list. list of categorical (str) labels for cluster composition of the piecharts (LHS subplot) length = n_samples.
    :param pie_size_scale: float default=0.8 scaling factor of the piechart nodes
    :return: f, ax, ax1
    '''

    from mpl_toolkits.axes_grid1 import make_axes_locatable
    f, ((ax, ax1)) = plt.subplots(1, 2, sharey=True,figsize=figsize, dpi=dpi)

    node_pos = via_object.graph_node_pos

    node_pos = np.asarray(node_pos)
    if cmap is None: cmap = 'coolwarm' if type_data == 'gene' else 'viridis_r'

    if type_data == 'pt':
        pt = via_object.scaled_hitting_times  # these are the final MCMC refined pt then slightly scaled at cluster level
        title_ax1 = "Pseudotime"

    if type_data == 'gene':
        pt = gene_exp
        title_ax1 = title
    if reference is None: reference_labels=via_object.true_label
    else: reference_labels = reference
    n_groups = len(set(via_object.labels))
    n_truegroups = len(set(reference_labels))
    group_pop = np.zeros([n_groups, 1])
    group_frac = pd.DataFrame(np.zeros([n_groups, n_truegroups]), columns=list(adata.obs[clusters].cat.categories))
    via_object.cluster_population_dict = {}
    for group_i in set(via_object.labels):
        loc_i = np.where(via_object.labels == group_i)[0]

        group_pop[group_i] = len(loc_i)  # np.sum(loc_i) / 1000 + 1
        via_object.cluster_population_dict[group_i] = len(loc_i)
        true_label_in_group_i = list(np.asarray(reference_labels)[loc_i])
        for ii in set(true_label_in_group_i):
            group_frac[ii][group_i] = true_label_in_group_i.count(ii)

    line_true = np.linspace(0, 1, n_truegroups)
    color_true_list = [plt.cm.rainbow(color) for color in line_true]
    color_true_list=adata.uns['{}_colors'.format(clusters)]

    sct = ax.scatter(node_pos[:, 0], node_pos[:, 1],
                     c='white', edgecolors='face', s=group_pop, cmap='jet')

    bboxes = getbb(sct, ax)

    ax = plot_edgebundle_viagraph(ax, via_object.hammerbundle_cluster, layout=via_object.graph_node_pos, CSM=via_object.CSM,
                            velocity_weight=via_object.velo_weight, pt=pt, headwidth_bundle=headwidth_arrow,
                            alpha_bundle=alpha_edge,linewidth_bundle=linewidth_edge, edge_color=edge_color)

    trans = ax.transData.transform
    bbox = ax.get_position().get_points()
    ax_x_min = bbox[0, 0]
    ax_x_max = bbox[1, 0]
    ax_y_min = bbox[0, 1]
    ax_y_max = bbox[1, 1]
    ax_len_x = ax_x_max - ax_x_min
    ax_len_y = ax_y_max - ax_y_min
    trans2 = ax.transAxes.inverted().transform
    pie_axs = []
    pie_size_ar = ((group_pop - np.min(group_pop)) / (np.max(group_pop) - np.min(group_pop)) + 0.5) / 10  # 10

    for node_i in range(n_groups):

        cluster_i_loc = np.where(np.asarray(via_object.labels) == node_i)[0]
        majority_true = via_object.func_mode(list(np.asarray(reference_labels)[cluster_i_loc]))
        pie_size = pie_size_ar[node_i][0] *pie_size_scale

        x1, y1 = trans(node_pos[node_i])  # data coordinates
        xa, ya = trans2((x1, y1))  # axis coordinates

        xa = ax_x_min + (xa - pie_size / 2) * ax_len_x
        ya = ax_y_min + (ya - pie_size / 2) * ax_len_y
        # clip, the fruchterman layout sometimes places below figure
        if ya < 0: ya = 0
        if xa < 0: xa = 0
        rect = [xa, ya, pie_size * ax_len_x, pie_size * ax_len_y]
        frac = np.asarray([ff for ff in group_frac.iloc[node_i].values])

        pie_axs.append(plt.axes(rect, frameon=False))
        pie_axs[node_i].pie(frac, wedgeprops={'linewidth': 0.0}, colors=color_true_list)
        pie_axs[node_i].set_xticks([])
        pie_axs[node_i].set_yticks([])
        pie_axs[node_i].set_aspect('equal')
        # pie_axs[node_i].text(0.5, 0.5, graph_node_label[node_i])
        if ax_text==True: pie_axs[node_i].text(0.5, 0.5, majority_true, fontsize = fontsize )

    patches, texts = pie_axs[node_i].pie(frac, wedgeprops={'linewidth': 0.0}, colors=color_true_list)
    labels = list(set(reference_labels))
    if show_legend ==True: plt.legend(patches, labels, loc=(-5, -5), fontsize=6, frameon=False)

    if via_object.time_series==True:
        ti = 'Cluster Composition. K=' + str(via_object.knn) + '. ncomp = ' + str(via_object.ncomp)  +'knnseq_'+str(via_object.knn_sequential)# "+ is_sub
    else:
        ti = 'Cluster Composition. K=' + str(via_object.knn) + '. ncomp = ' + str(via_object.ncomp)
    ax.set_title(ti)
    ax.grid(False)
    ax.set_xticks([])
    ax.set_yticks([])

    title_list = [title_ax1]
    for i, ax_i in enumerate([ax1]):
        pt = via_object.markov_hitting_times if type_data == 'pt' else gene_exp

        c_edge, l_width = [], []
        for ei, pti in enumerate(pt):
            if ei in via_object.terminal_clusters:
                c_edge.append('red')
                l_width.append(1.5)
            else:
                c_edge.append('gray')
                l_width.append(0.0)

        gp_scaling = 1000 / max(group_pop)

        group_pop_scale = group_pop * gp_scaling * 0.5
        ax_i=plot_edgebundle_viagraph(ax_i, via_object.hammerbundle_cluster, layout=via_object.graph_node_pos,CSM=via_object.CSM, velocity_weight=via_object.velo_weight, pt=pt,headwidth_bundle=headwidth_arrow, alpha_bundle=alpha_edge, linewidth_bundle=linewidth_edge, edge_color=edge_color)

        im1 = ax_i.scatter(node_pos[:, 0], node_pos[:, 1], s=group_pop_scale, c=pt, cmap=cmap,
                           edgecolors=c_edge,
                           alpha=1, zorder=3, linewidth=l_width)
        if ax_text:
            x_max_range = np.amax(node_pos[:, 0]) / 100
            y_max_range = np.amax(node_pos[:, 1]) / 100

            for ii in range(node_pos.shape[0]):
                ax_i.text(node_pos[ii, 0] + max(x_max_range, y_max_range),
                          node_pos[ii, 1] + min(x_max_range, y_max_range),
                          'C' + str(ii) + 'pop' + str(int(group_pop[ii][0])),
                          color='black', zorder=4, fontsize = fontsize)
        ax_i.set_title(title_list[i])
        ax_i.grid(False)
        ax_i.set_xticks([])
        ax_i.set_yticks([])

    divider = make_axes_locatable(ax1)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    if type_data == 'pt':
        f.colorbar(im1, cax=cax, orientation='vertical', label='pseudotime',shrink=0.5)
    else:
        f.colorbar(im1, cax=cax, orientation='vertical', label='Gene expression',shrink=0.5)
    f.patch.set_visible(False)

    ax1.axis('off')
    ax.axis('off')
    ax.set_facecolor('white')
    ax1.set_facecolor('white')
    return f, ax, ax1


def get_gene_expression_pyomic(via0, gene_exp:pd.DataFrame, cmap:str='jet', figsize=(8,4),
                        dpi:int=150, marker_genes:list = [], linewidth:float = 2.0,
                        n_splines:int=10, spline_order:int=4, fontsize_:int=10, marker_lineages=[]):
    '''
    :param gene_exp: dataframe where columns are features (gene) and rows are single cells
    :param cmap: default: 'jet'
    :param dpi: default:150
    :param marker_genes: Default is to use all genes in gene_exp. other provide a list of marker genes that will be used from gene_exp.
    :param linewidth: default:2
    :param n_slines: default:10
    :param spline_order: default:4
    :param marker_lineages: Default is to use all lineage pathways. other provide a list of lineage number (terminal cluster number).
    :return: fig, axs
    '''

    if len(marker_lineages)==0: marker_lineages=via0.terminal_clusters

    if len(marker_genes) >0: gene_exp=gene_exp[marker_genes]
    sc_pt = via0.single_cell_pt_markov
    sc_bp_original = via0.single_cell_bp
    n_terminal_states = sc_bp_original.shape[1]

    palette = cm.get_cmap(cmap, n_terminal_states)
    cmap_ = palette(range(n_terminal_states))
    n_genes = gene_exp.shape[1]

    fig_nrows, mod = divmod(n_genes, 4)
    if mod == 0: fig_nrows = fig_nrows
    if mod != 0: fig_nrows += 1

    fig_ncols = 4
    fig, axs = plt.subplots(fig_nrows, fig_ncols, dpi=dpi, figsize=figsize)
    fig.patch.set_visible(False)
    i_gene = 0  # counter for number of genes
    i_terminal = 0 #counter for terminal cluster
    # for i in range(n_terminal_states): #[0]

    for r in range(fig_nrows):
        for c in range(fig_ncols):
            if (i_gene < n_genes):
                for i_terminal in range(n_terminal_states):
                    sc_bp = sc_bp_original.copy()
                    if (via0.terminal_clusters[i_terminal] in marker_lineages and len(np.where(sc_bp[:, i_terminal] > 0.9)[ 0]) > 0): # check if terminal state is in marker_lineage and in case this terminal state i cannot be reached (sc_bp is all 0)
                        gene_i = gene_exp.columns[i_gene]
                        loc_i = np.where(sc_bp[:, i_terminal] > 0.9)[0]
                        val_pt = [sc_pt[pt_i] for pt_i in loc_i]  # TODO,  replace with array to speed up

                        max_val_pt = max(val_pt)

                        loc_i_bp = np.where(sc_bp[:, i_terminal] > 0.000)[0]  # 0.001
                        loc_i_sc = np.where(np.asarray(sc_pt) <= max_val_pt)[0]

                        loc_ = np.intersect1d(loc_i_bp, loc_i_sc)

                        gam_in = np.asarray(sc_pt)[loc_]
                        x = gam_in.reshape(-1, 1)

                        y = np.asarray(gene_exp[gene_i])[loc_].reshape(-1, 1)

                        weights = np.asarray(sc_bp[:, i_terminal])[loc_].reshape(-1, 1)

                        if len(loc_) > 1:
                            geneGAM = pg.LinearGAM(n_splines=n_splines, spline_order=spline_order, lam=10).fit(x, y, weights=weights)
                            xval = np.linspace(min(sc_pt), max_val_pt, 100 * 2)
                            yg = geneGAM.predict(X=xval)

                        else:
                            print(f'{datetime.now()}\tLineage {i_terminal} cannot be reached. Exclude this lineage in trend plotting')

                        if fig_nrows >1:
                            axs[r,c].plot(xval, yg, color=cmap_[i_terminal], linewidth=linewidth, zorder=3, label=f"Lineage:{via0.terminal_clusters[i_terminal]}")
                            axs[r,c].set_title(gene_i,fontsize=fontsize_)
                            # Set tick font size
                            for label in (axs[r,c].get_xticklabels() + axs[r,c].get_yticklabels()):
                                label.set_fontsize(fontsize_-1)
                            #if i_gene == n_genes -1:
                            #    axs[r,c].legend(frameon=False, fontsize=fontsize_)
                            #    axs[r, c].set_xlabel('Time', fontsize=fontsize_)
                            #    axs[r, c].set_ylabel('Intensity', fontsize=fontsize_)
                            axs[r,c].spines['top'].set_visible(False)
                            axs[r,c].spines['right'].set_visible(False)
                            axs[r,c].legend().set_visible(False)
                            axs[r,c].grid(False)
                        else:
                            axs[c].plot(xval, yg, color=cmap_[i_terminal], linewidth=linewidth, zorder=3,   label=f"Lineage:{via0.terminal_clusters[i_terminal]}")
                            axs[c].set_title(gene_i, fontsize=fontsize_)
                            # Set tick font size
                            for label in (axs[c].get_xticklabels() + axs[c].get_yticklabels()):
                                label.set_fontsize(fontsize_-1)
                            #if i_gene == n_genes -1:
                            #    axs[c].legend(frameon=False,fontsize=fontsize_)
                            #    axs[c].set_xlabel('Time', fontsize=fontsize_)
                            #    axs[c].set_ylabel('Intensity', fontsize=fontsize_)
                            axs[c].spines['top'].set_visible(False)
                            axs[c].spines['right'].set_visible(False)
                            axs[c].legend().set_visible(False)
                            axs[c].grid(False)
                i_gene+=1
            else:
                if fig_nrows > 1:
                    axs[r,c].axis('off')
                    axs[r, c].grid(False)
                else:
                    axs[c].axis('off')
                    axs[c].grid(False)
    

    handles, labels = axs[0,0].get_legend_handles_labels()
    fig.legend(handles, labels, loc='lower center', ncol=fig_ncols, bbox_to_anchor=(0.5, -0.25))
    # 添加全局横纵标签
    #fig.suptitle('My Figure Title', fontsize=16, fontweight='bold')
    fig.text(0.5, 0.03, 'Time', ha='center', fontsize=fontsize_+2)
    fig.text(0.04, 0.5, 'Intensity', va='center', rotation='vertical', fontsize=fontsize_+2)


    return fig, axs

def draw_clustergraph_pyomic(via_object, type_data='gene', gene_exp='', gene_list='', arrow_head=0.1,
                      edgeweight_scale=1.5, cmap=None, label_=True,figsize=(8,4),dpi=80):
    '''
    :param via_object:
    :param type_data:
    :param gene_exp:
    :param gene_list:
    :param arrow_head:
    :param edgeweight_scale:
    :param cmap:
    :param label_:
    :return: fig, axs
    '''
    '''
    #draws the clustergraph for cluster level gene or pseudotime values
    # type_pt can be 'pt' pseudotime or 'gene' for gene expression
    # ax1 is the pseudotime graph
    '''
    n = len(gene_list)
    fig_nrows, mod = divmod(n, 4)
    if mod == 0: fig_nrows = fig_nrows
    if mod != 0: fig_nrows += 1

    fig_ncols = 4
    fig, axs = plt.subplots(fig_nrows, fig_ncols, dpi=dpi, figsize=figsize)
    pt = via_object.markov_hitting_times
    if cmap is None: cmap = 'coolwarm' if type_data == 'gene' else 'viridis_r'

    node_pos = via_object.graph_node_pos
    edgelist = list(via_object.edgelist_maxout)
    edgeweight = via_object.edgeweights_maxout

    node_pos = np.asarray(node_pos)

    import matplotlib.lines as lines


    n_groups = len(set(via_object.labels))  # node_pos.shape[0]
    n_truegroups = len(set(via_object.true_label))
    group_pop = np.zeros([n_groups, 1])
    via_object.cluster_population_dict = {}
    for group_i in set(via_object.labels):
        loc_i = np.where(via_object.labels == group_i)[0]

        group_pop[group_i] = len(loc_i)  # np.sum(loc_i) / 1000 + 1
        via_object.cluster_population_dict[group_i] = len(loc_i)

    for i in range(n):
        ax_i = axs[i]
        gene_i = gene_list[i]
        '''
        for e_i, (start, end) in enumerate(edgelist):
            if pt[start] > pt[end]:
                start, end = end, start

            ax_i.add_line(
                lines.Line2D([node_pos[start, 0], node_pos[end, 0]], [node_pos[start, 1], node_pos[end, 1]],
                             color='black', lw=edgeweight[e_i] * edgeweight_scale, alpha=0.5))
            z = np.polyfit([node_pos[start, 0], node_pos[end, 0]], [node_pos[start, 1], node_pos[end, 1]], 1)
            minx = np.min(np.array([node_pos[start, 0], node_pos[end, 0]]))

            direction = 1 if node_pos[start, 0] < node_pos[end, 0] else -1
            maxx = np.max([node_pos[start, 0], node_pos[end, 0]])
            xp = np.linspace(minx, maxx, 500)
            p = np.poly1d(z)
            smooth = p(xp)
            step = 1

            ax_i.arrow(xp[250], smooth[250], xp[250 + direction * step] - xp[250],
                       smooth[250 + direction * step] - smooth[250],
                       shape='full', lw=0, length_includes_head=True, head_width=arrow_head_w, color='grey')
        '''
        c_edge, l_width = [], []
        for ei, pti in enumerate(pt):
            if ei in via_object.terminal_clusters:
                c_edge.append('red')
                l_width.append(1.5)
            else:
                c_edge.append('gray')
                l_width.append(0.0)
        ax_i = plot_edgebundle_viagraph(ax_i, via_object.hammerbundle_cluster, layout=via_object.graph_node_pos, CSM=via_object.CSM,
                                velocity_weight=via_object.velo_weight, pt=pt, headwidth_bundle=arrow_head, alpha_bundle=0.4, linewidth_bundle=edgeweight_scale)
        group_pop_scale = .5 * group_pop * 1000 / max(group_pop)
        pos = ax_i.scatter(node_pos[:, 0], node_pos[:, 1], s=group_pop_scale, c=gene_exp[gene_i].values, cmap=cmap,
                           edgecolors=c_edge, alpha=1, zorder=3, linewidth=l_width)
        if label_==True:
            for ii in range(node_pos.shape[0]):
                ax_i.text(node_pos[ii, 0] + 0.1, node_pos[ii, 1] + 0.1, 'C'+str(ii)+' '+str(round(gene_exp[gene_i].values[ii], 1)),
                          color='black', zorder=4, fontsize=6)
        divider = make_axes_locatable(ax_i)
        cax = divider.append_axes('right', size='10%', pad=0.05)

        cbar=fig.colorbar(pos, cax=cax, orientation='vertical')
        cbar.ax.tick_params(labelsize=8)
        ax_i.set_title(gene_i)
        ax_i.grid(False)
        ax_i.set_xticks([])
        ax_i.set_yticks([])
        ax_i.axis('off')
    fig.patch.set_visible(False)
    return fig, axs

def plot_gene_trend_heatmaps_pyomic(via_object, df_gene_exp:pd.DataFrame, marker_lineages:list = [], 
                             fontsize:int=8,cmap:str='viridis', normalize:bool=True, ytick_labelrotation:int = 0, 
                             figsize:tuple=(2,4)):
    '''

    Plot the gene trends on heatmap: a heatmap is generated for each lineage (identified by terminal cluster number). Default selects all lineages

    :param via_object:
    :param df_gene_exp: pandas DataFrame single-cell level expression [cells x genes]
    :param marker_lineages: list default = None and plots all detected all lineages. Optionally provide a list of integers corresponding to the cluster number of terminal cell fates
    :param fontsize: int default = 8
    :param cmap: str default = 'viridis'
    :param normalize: bool = True
    :param ytick_labelrotation: int default = 0
    :return: fig and list of axes
    '''
    import seaborn as sns

    if len(marker_lineages) ==0: marker_lineages = via_object.terminal_clusters
    dict_trends = get_gene_trend(via_object=via_object, marker_lineages=marker_lineages, df_gene_exp=df_gene_exp)
    branches = list(dict_trends.keys())
    genes = dict_trends[branches[0]]['trends'].index
    height = len(genes) * len(branches)
    # Standardize the matrix (standardization along each gene. Since SS function scales the columns, we first transpose the df)
    #  Set up plot
    fig = plt.figure(figsize=figsize)
    ax_list = []
    for i, branch in enumerate(branches):
        ax = fig.add_subplot(len(branches), 1, i + 1)
        df_trends=dict_trends[branch]['trends']
        # normalize each genes (feature)
        if normalize==True:
            df_trends = pd.DataFrame(
            StandardScaler().fit_transform(df_trends.T).T,
            index=df_trends.index,
            columns=df_trends.columns)

        ax.set_title('Lineage: ' + str(branch) + '-' + str(dict_trends[branch]['name']), fontsize=int(fontsize*1.3))
        #sns.set(size=fontsize)  # set fontsize 2
        b=sns.heatmap(df_trends,yticklabels=True, xticklabels=False, cmap = cmap)
        b.tick_params(labelsize=fontsize,labelrotation=ytick_labelrotation)
        b.figure.axes[-1].tick_params(labelsize=fontsize)
        ax_list.append(ax)
    b.set_xlabel("pseudotime", fontsize=int(fontsize*1.3))
    return fig, ax_list




import torch
from torch_geometric.loader import DataLoader
from tqdm import tqdm
import torch.nn.functional as F
from ..STAGATE_pyG import Batch_Data,Cal_Spatial_Net,Transfer_pytorch_Data,Stats_Spatial_Net,STAGATE
class pySTAGATE(object):
    
    def __init__(self,adata,num_batch_x,num_batch_y,
                 spatial_key=['X','Y'],batch_size=1,
                rad_cutoff=200,num_epoch = 1000,lr=0.001,
                weight_decay=1e-4,hidden_dims = [512, 30],
                device='cuda:0'):
        
        device = torch.device(device if torch.cuda.is_available() else 'cpu')
        self.device=device
        
        Batch_list = Batch_Data(adata, num_batch_x=num_batch_x, num_batch_y=num_batch_y,
                                    spatial_key=spatial_key, plot_Stats=True)
        for temp_adata in Batch_list:
            Cal_Spatial_Net(temp_adata, rad_cutoff=rad_cutoff)
        
        
        #device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        data_list = [Transfer_pytorch_Data(adata) for adata in Batch_list]
        for temp in data_list:
            temp.to(device)
        
        Cal_Spatial_Net(adata, rad_cutoff=rad_cutoff)
        data = Transfer_pytorch_Data(adata)
        Stats_Spatial_Net(adata)
        # batch_size=1 or 2
        self.loader = DataLoader(data_list, batch_size=batch_size, shuffle=True)
        
        # hyper-parameters
        self.num_epoch = num_epoch
        self.lr=lr
        self.weight_decay=weight_decay
        self.hidden_dims = hidden_dims
        self.adata=adata
        self.data=data
        
        self.model = STAGATE(hidden_dims = [data_list[0].x.shape[1]]+self.hidden_dims).to(device)

        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr, weight_decay=self.weight_decay)

        
    def train(self):
        for epoch in tqdm(range(1, self.num_epoch+1)):
            for batch in self.loader:
                self.model.train()
                self.optimizer.zero_grad()
                z, out = self.model(batch.x, batch.edge_index)
                loss = F.mse_loss(batch.x, out) #F.nll_loss(out[data.train_mask], data.y[data.train_mask])
                loss.backward()
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), 5.)
                self.optimizer.step()
        # The total network
        self.data.to(self.device)
        
    def predicted(self):
        self.model.eval()
        z, out = self.model(self.data.x, self.data.edge_index)

        STAGATE_rep = z.to('cpu').detach().numpy()
        self.adata.obsm['STAGATE'] = STAGATE_rep
        ReX = out.to('cpu').detach().numpy()
        ReX[ReX<0] = 0
        self.adata.layers['STAGATE_ReX'] = ReX
        


import numpy as np
import pandas as pd
import os.path

import scanpy as sc
from scipy.io import mmread


def _clean_up(df):
    df = df.loc[df.index[df.sum(axis=1) > 0], :]
    df = df.loc[:, df.columns[df.sum() > 0]]
    return df


def from_csv(counts_csv_file, delimiter=","):
    # Read in csv file
    df = pd.read_csv(counts_csv_file, sep=delimiter, index_col=0)
    clean_df = _clean_up(df)
    return clean_df


def from_mtx(mtx_file, gene_name_file):
    # Read in mtx file
    count_matrix = mmread(mtx_file)

    gene_names = np.loadtxt(gene_name_file, dtype=np.dtype("S"))
    gene_names = np.array([gene.decode("utf-8") for gene in gene_names])

    # remove todense
    df = pd.DataFrame(count_matrix.todense(), columns=gene_names)

    return _clean_up(df)


def from_10x(data_dir, use_ensemble_id=True):
    # loads 10x sparse format data
    # data_dir is dir that contains matrix.mtx, genes.tsv and barcodes.tsv
    # return_sparse=True -- returns data matrix in sparse format (default = False)

    if data_dir is None:
        data_dir = "./"
    elif data_dir[len(data_dir) - 1] != "/":
        data_dir = data_dir + "/"

    filename_dataMatrix = os.path.expanduser(data_dir + "matrix.mtx")
    filename_genes = os.path.expanduser(data_dir + "genes.tsv")
    filename_cells = os.path.expanduser(data_dir + "barcodes.tsv")

    # Read in gene expression matrix (sparse matrix)
    # Rows = genes, columns = cells
    dataMatrix = mmread(filename_dataMatrix)

    # Read in row names (gene names / IDs)
    gene_names = np.loadtxt(filename_genes, delimiter="\t", dtype=bytes).astype(str)
    if use_ensemble_id:
        gene_names = [gene[0] for gene in gene_names]
    else:
        gene_names = [gene[1] for gene in gene_names]
    cell_names = np.loadtxt(filename_cells, delimiter="\t", dtype=bytes).astype(str)

    dataMatrix = pd.DataFrame(
        dataMatrix.todense(), columns=cell_names, index=gene_names
    )

    # combine duplicate genes
    if not use_ensemble_id:
        dataMatrix = dataMatrix.groupby(dataMatrix.index).sum()
    dataMatrix = dataMatrix.transpose()

    return _clean_up(dataMatrix)


def from_10x_HDF5(filename, genome=None):
    ad = sc.read_10x_h5(filename, genome, True)

    dataMatrix = pd.DataFrame(ad.X.todense(), columns=ad.var_names, index=ad.obs_names)

    return _clean_up(dataMatrix)


def from_fcs(
    cls,
    fcs_file,
    cofactor=5,
    metadata_channels=[
        "Time",
        "Event_length",
        "DNA1",
        "DNA2",
        "Cisplatin",
        "beadDist",
        "bead1",
    ],
):
    # Parse the fcs file
    import fcsparser
    text, data = fcsparser.parse(fcs_file)
    data = data.astype(np.float64)

    # Extract the S and N features (Indexing assumed to start from 1)
    # Assumes channel names are in S
    no_channels = text["$PAR"]
    channel_names = [""] * no_channels
    for i in range(1, no_channels + 1):
        # S name
        try:
            channel_names[i - 1] = text["$P%dS" % i]
        except KeyError:
            channel_names[i - 1] = text["$P%dN" % i]
    data.columns = channel_names

    # Metadata and data
    metadata_channels = data.columns.intersection(metadata_channels)
    data_channels = data.columns.difference(metadata_channels)
    # metadata = data[metadata_channels]
    data = data[data_channels]

    # Transform if necessary
    if cofactor is not None or cofactor > 0:
        data = np.arcsinh(np.divide(data, cofactor))

    return data



import numpy as np
import scipy as s
import scipy.special as special
from .basic_distributions import Distribution

from ..utils import *

class Beta(Distribution):
    """
    Class to define Beta distributions

    Equations:
    p(x|a,b) = GammaF(a+b)/(GammaF(a)*GammaF(b)) * x**(a-1) * (1-x)**(b-1)
    log p(x|a,b) = log[GammaF(a+b)/(GammaF(a)*GammaF(b))] + (a-1)*x + (b-1)*(1-x)
    E[x] = a/(a+b)
    var[x] = a*b / ((a+b)**2 * (a+b+1))
    """
    def __init__(self, dim, a, b, E=None):
        Distribution.__init__(self, dim)

        # Initialise parameters
        a = np.ones(dim)*a
        b = np.ones(dim)*b
        self.params = { 'a':a, 'b':b }

        # Initialise expectations
        if E is None:
            self.updateExpectations()
        else:
            self.expectations = {
               'E': np.ones(dim) * E,
               'lnE': np.log(np.ones(dim) * E),
               'lnEInv': np.log(1. - np.ones(dim) * E)
            }
            self.expectations["lnEInv"][np.isinf(self.expectations["lnEInv"])] = -np.inf
            # self.updateExpectations()
            # print("The expectation of the Beta distribution is initialized consistently with the provided parameters (not with the provided expectation)")

        # Check that dimensionalities match
        self.CheckDimensionalities()

    def updateExpectations(self):
        a, b = self.params['a'], self.params['b']
        E = np.divide(a,a+b)
        lnE = special.digamma(a) - special.digamma(a+b)
        lnEInv = special.digamma(b) - special.digamma(a+b) # expectation of ln(1-X)
        lnEInv[np.isinf(lnEInv)] = -np.inf # there is a numerical error in lnEInv if E=1
        self.expectations = { 'E':E, 'lnE':lnE, 'lnEInv':lnEInv }

    def sample(self, n=1):
        a = self.params['a']
        b = self.params['b']
        return np.random.beta(a, b)
